"""
Global Offset Table Tracker

Subsystem for tracking accesses to external function calls made through pointers
in an inferior's Global Offset Table, such as those made by the stubs in the
Procedure Linkage Table.

Currently, it does this by attatching watchpoints to the entries in the GOT and
taking note of where the call came from, but it could be done much faster by
injecting our own code into the program space to track this.
"""

from __future__ import annotations

import gdb

import pwndbg.aglib
import pwndbg.aglib.dynamic
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.color.message as message
import pwndbg.gdblib.bpoint
import pwndbg.gdblib.shellcode
import pwndbg.lib.memory


class RelocTypes:
    """
    This class contains all the relocation type constants so that one may
    interpret the relocations types present in the DYNAMIC segment. These
    constants are defined in each of the processors' SystemV R4 psABI document,
    or equivalent, and should stay the same across all implementations of libc
    on systems that adhere to that ABI, such as Linux.

    Most of these were sourced from GLibc, which conveniently lists all of the
    relocations types in a single file[1].

    [1]: https://elixir.bootlin.com/glibc/glibc-2.38/source/elf/elf.h
    """

    R_RISCV_JUMP_SLOT = 5
    R_X86_64_JUMP_SLOT = 7
    R_386_JMP_SLOT = 7
    R_CRIS_JUMP_SLOT = 11
    R_390_JMP_SLOT = 11
    R_CKCORE_JUMP_SLOT = 12
    R_TILEPRO_JMP_SLOT = 12
    R_MICROBLAZE_JUMP_SLOT = 17
    R_TILEGX_JMP_SLOT = 18
    R_OR1K_JMP_SLOT = 20
    R_68K_JMP_SLOT = 21
    R_SPARC_JMP_SLOT = 21
    R_PPC_JMP_SLOT = 21
    R_PPC64_JMP_SLOT = 21
    R_ARM_JUMP_SLOT = 22
    R_MN10300_JMP_SLOT = 22
    R_ALPHA_JMP_SLOT = 26
    R_NIOS2_JUMP_SLOT = 38
    R_NDS32_JMP_SLOT = 41
    R_METAG_JMP_SLOT = 44
    R_M32R_JMP_SLOT = 52
    R_ARC_JMP_SLOT = 55
    R_MIPS_JUMP_SLOT = 127
    R_SH_JMP_SLOT = 164
    R_AARCH64_JUMP_SLOT = 1026

    R_X86_64_IRELATIVE = 37
    R_386_IRELATIVE = 42
    R_RISCV_IRELATIVE = 58
    R_390_IRELATIVE = 61
    R_ARM_IRELATIVE = 160
    R_AARCH64_P32_IRELATIVE = 188
    R_PPC_IRELATIVE = 248
    R_PPC64_IRELATIVE = 248
    R_SPARC_IRELATIVE = 249
    R_AARCH64_IRELATIVE = 1032


# Set of all type codes associated with jump slots, by architecture.
JUMP_SLOTS = {
    "x86-64": {RelocTypes.R_X86_64_JUMP_SLOT},
    "i386": {RelocTypes.R_386_JMP_SLOT},
    "aarch64": {RelocTypes.R_AARCH64_JUMP_SLOT},
    "mips": {RelocTypes.R_MIPS_JUMP_SLOT},
    "powerpc": {RelocTypes.R_PPC_JMP_SLOT},
    "sparc": {RelocTypes.R_SPARC_JMP_SLOT},
    "arm": {RelocTypes.R_ARM_JUMP_SLOT},
    "armcm": {RelocTypes.R_ARM_JUMP_SLOT},
    "rv32": {RelocTypes.R_RISCV_JUMP_SLOT},
    "rv64": {RelocTypes.R_RISCV_JUMP_SLOT},
}

# Set of all type codes associated with irelative jump slots, by architecture.
IRELATIVE_SLOTS = {
    "x86-64": {RelocTypes.R_X86_64_IRELATIVE},
    "i386": {RelocTypes.R_386_IRELATIVE},
    "aarch64": {RelocTypes.R_AARCH64_P32_IRELATIVE, RelocTypes.R_AARCH64_IRELATIVE},
    "mips": set(),
    "powerpc": {RelocTypes.R_PPC_IRELATIVE},
    "sparc": {RelocTypes.R_SPARC_IRELATIVE},
    "arm": {RelocTypes.R_ARM_IRELATIVE},
    "armcm": {RelocTypes.R_ARM_IRELATIVE},
    "rv32": {RelocTypes.R_RISCV_IRELATIVE},
    "rv64": {RelocTypes.R_RISCV_IRELATIVE},
}


def is_mmap_error(ptr: int):
    """
    Checks whether the return value of an mmap of indicates an error.
    """
    err = ((1 << pwndbg.aglib.arch.ptrsize) - 1) & pwndbg.lib.memory.PAGE_MASK
    return ptr & pwndbg.lib.memory.PAGE_MASK == err


class TrapAllocator:
    """
    Utility that allocates and manages executable addresses in the space of the
    executing program that we can trap.
    """

    block_capacity = 4096
    slot_size = 8

    def __init__(self) -> None:
        self._reset()

    def _reset(self) -> None:
        """
        Reset the internal state of the allocator.
        """
        self.blocks: list[int] = []
        self.current_block_occupancy = 0
        self.vacant_slots: list[int] = []
        self.occupied_slots: set[int] = set()

    def alloc(self):
        """
        Allocates a new address to where program execution can be diverted.
        """
        if len(self.vacant_slots) > 0:
            # We have an easy vacant slot we can recycle.
            addr = self.vacant_slots.pop()
            self.occupied_slots.add(addr)
            return addr

        if len(self.blocks) > 0 and self.current_block_occupancy < self.block_capacity:
            # We have a non-full block we can allocate a new slot from.
            addr = self.blocks[-1] + self.current_block_occupancy * self.slot_size
            self.current_block_occupancy += 1
            self.occupied_slots.add(addr)
            return addr

        # We have to allocate a new block.
        block_base = pwndbg.gdblib.shellcode.exec_syscall(
            "SYS_mmap",
            0,
            self.block_capacity * self.slot_size,
            5,  # PROT_READ | PROT_EXEC
            0x22,  # MAP_PRIVATE | MAP_ANONYMOUS
            -1,
            0,
            disable_breakpoints=True,
        )
        if is_mmap_error(block_base):
            raise RuntimeError(f"SYS_mmap request returned {block_base:#x}")

        self.blocks.append(block_base)
        self.current_block_occupancy = 1
        addr = self.blocks[-1]

        self.occupied_slots.add(addr)
        return addr

    def free(self, address) -> None:
        """
        Indicates that an address obtained from alloc() can be recycled.
        """
        assert address in self.occupied_slots
        self.occupied_slots.remove(address)
        self.vacant_slots.append(address)

    def clear(self):
        """
        Deletes all memory mappings and frees all addresses.
        """
        size = self.block_capacity * self.slot_size
        while len(self.blocks) > 0:
            block = self.blocks.pop()
            result = pwndbg.gdblib.shellcode.exec_syscall(
                "SYS_munmap", block, size, disable_breakpoints=True
            )
            if result != 0:
                raise RuntimeError(f"SYS_munmap({block:#x}, {size:#x}) failed ({result:#x})")
        self._reset()


def display_name(name, basename=False):
    """
    Return the display name for a symbol or objfile.

    Ideally, we'd like to display all of the names of the symbols as text, but
    there is really nothing stopping symbol names from being stored in some
    fairly wacky encoding or really from having names that aren't text at all.

    We should try our best to turn whatever the symbol name is into text, but
    not so much that non-text entries or entries in unknown encodings become
    unrecognizable.
    """
    if name == b"":
        return "<Empty>"
    try:
        if isinstance(name, bytearray):
            name = name.decode("ascii")
        if basename and "/" in name:
            name = name.split("/")[-1]
        return name
    except TypeError:
        return name


# The allocator we use for our trap addresses.
TRAP_ALLOCATOR = TrapAllocator()

# Whether the GOT tracking is currently enabled.
GOT_TRACKING = False

# Map describing all of the currently installed analysis watchpoints.
INSTALLED_WATCHPOINTS: dict[int, tuple[Tracker, Patcher]] = {}


class Patcher(pwndbg.gdblib.bpoint.Breakpoint):
    """
    Watches for changes made by program code to the GOT and fixes them up.

    This class is paired with Tracker, and instances of both classes always
    function together.
    """

    entry = 0
    tracker = None

    def __init__(self, entry, tracker) -> None:
        super().__init__(
            f"*(void**){entry:#x}", type=gdb.BP_WATCHPOINT, wp_class=gdb.WP_WRITE, internal=True
        )
        self.silent = True
        self.entry = entry
        self.tracker = tracker
        self.init = True

        # Figure out the display names both this class and its corresponding
        # tracker will use.
        objfile = self.tracker.link_map_entry.name()
        if objfile == b"":
            objfile = pwndbg.aglib.proc.exe()
        self.tracker.obj_display_name = display_name(objfile, basename=True)

        self.tracker.sym_display_name = display_name(
            self.tracker.dynamic_section.string(
                self.tracker.dynamic_section.symtab_read(
                    self.tracker.relocation_fn(self.tracker.relocation_index, "r_sym"), "st_name"
                )
            )
        )

    def should_stop(self) -> bool:
        # Read the new branch target, and update the redirection target of the
        # tracker accordingly.
        new_target = pwndbg.aglib.memory.read_pointer_width(self.entry)
        if new_target == self.tracker.trapped_address:
            # The write to this range from within GDB that we do at the end of
            # this function can cause this watchpoint to trigger again.
            # Obviously, we don't want to treat our own writes the same way we'd
            # treat writes made by the inferior.
            return False

        self.tracker.target = new_target

        # Notify the user about changes to the GOT.
        if not self.init:
            print(
                f"[*] GOT entry {self.entry:#x} ({self.tracker.sym_display_name}@{self.tracker.obj_display_name}) now points to {new_target:#x}"
            )
        self.init = False

        # Update the GOT entry so that it points to the trapped address again.
        #
        # FIXME: Ideally, we'd use gdb.Value([...]).assign() here, but that is
        # not always available, so we must do this ugly hack instead.
        gdb.execute(f"set *(void**){self.entry:#x} = {self.tracker.trapped_address:#x}")

        return False


class Tracker(pwndbg.gdblib.bpoint.Breakpoint):
    """
    Class that tracks the accesses made to the entries in the GOT.

    This class is paired with Patcher, and instances of both classes always
    function together.
    """

    hits: dict[tuple[int, ...], int] = {}
    total_hits = 0
    trapped_address = 0

    target = 0
    dynamic_section = None
    relocation_fn = None
    relocation_index = 0
    link_map_entry = None

    def __init__(self) -> None:
        self.trapped_address = TRAP_ALLOCATOR.alloc()
        super().__init__(f"*{self.trapped_address:#x}", internal=True)
        self.hits = {}
        self.silent = True

    def delete(self) -> None:
        TRAP_ALLOCATOR.free(self.trapped_address)
        super().delete()

    def should_stop(self) -> bool:
        # Notify the user about calls made through this GOT entry.
        print(f"[*] {self.sym_display_name}@{self.obj_display_name} called via GOT")

        # Collect the stack that accessed this GOT entry.
        stack = []
        frame = gdb.newest_frame().older()
        while frame is not None:
            stack.append(int(frame.pc()))
            frame = frame.older()
        stack = tuple(stack)
        if stack not in self.hits:
            self.hits[stack] = 0
        self.hits[stack] += 1
        self.total_hits += 1

        # Divert execution back to the real jump target.
        gdb.execute(f"set $pc = {self.target}")
        return False


def _update_watchpoints() -> None:
    """
    Internal function responsible for updating the watchpoints that track the
    accesses to the GOT.
    """
    if not GOT_TRACKING:
        # We don't want to bother anyone.
        return

    # Remove the watchpoints that are currently enabled.
    for _, (tracker, patcher) in INSTALLED_WATCHPOINTS.items():
        patcher.delete()
        tracker.delete()
    INSTALLED_WATCHPOINTS.clear()

    # Install new watchpoints to cover all of the jump slots in all GOTs
    for obj in pwndbg.aglib.dynamic.link_map():
        name = obj.name()
        if name == b"":
            name = pwndbg.aglib.proc.exe()

        try:
            dynamic = pwndbg.aglib.dynamic.DynamicSegment(obj.dynamic(), obj.load_bias())
        except RuntimeError as e:
            print(message.warn(f"object {name} has invalid DYNAMIC section: {e}"))
            continue

        jump_slots = JUMP_SLOTS[pwndbg.aglib.arch.name]
        if dynamic.has_rel:
            for i in range(dynamic.rel_entry_count()):
                if dynamic.rel_read(i, "r_type") not in jump_slots:
                    continue
                target = dynamic.load_bias + dynamic.rel_read(i, "r_offset")

                tracker = Tracker()
                tracker.dynamic_section = dynamic
                tracker.link_map_entry = obj
                tracker.relocation_index = i
                tracker.relocation_fn = dynamic.rel_read
                patcher = Patcher(target, tracker)
                patcher.should_stop()

                INSTALLED_WATCHPOINTS[target] = (tracker, patcher)
        if dynamic.has_rela:
            for i in range(dynamic.rela_entry_count()):
                if dynamic.rela_read(i, "r_type") not in jump_slots:
                    continue
                target = (
                    dynamic.load_bias
                    + dynamic.rela_read(i, "r_offset")
                    + dynamic.rela_read(i, "r_addend")
                )

                tracker = Tracker()
                tracker.dynamic_section = dynamic
                tracker.link_map_entry = obj
                tracker.relocation_index = i
                tracker.relocation_fn = dynamic.rela_read
                patcher = Patcher(target, tracker)
                patcher.should_stop()

                INSTALLED_WATCHPOINTS[target] = (tracker, patcher)
        if dynamic.has_jmprel:
            for i in range(dynamic.jmprel_entry_count()):
                if dynamic.jmprel_read(i, "r_type") not in jump_slots:
                    continue
                target = dynamic.load_bias + dynamic.jmprel_read(i, "r_offset")
                if dynamic.jmprel_elem.has_field("r_addend"):
                    target += dynamic.jmprel_read(i, "r_addend")

                tracker = Tracker()
                tracker.dynamic_section = dynamic
                tracker.link_map_entry = obj
                tracker.relocation_index = i
                tracker.relocation_fn = dynamic.jmprel_read
                patcher = Patcher(target, tracker)
                patcher.should_stop()

                INSTALLED_WATCHPOINTS[target] = (tracker, patcher)


# Set the function so that it's called whenever the link map changes.
pwndbg.aglib.dynamic.r_debug_link_map_changed_add_listener(_update_watchpoints)


def all_tracked_entries():
    """
    Return an iterator over all of the GOT whose accesses are being tracked.
    """
    return INSTALLED_WATCHPOINTS.items()


def tracked_entry_by_address(address):
    """
    Return the tracker associated with the entry at the given address, if any.
    """
    return INSTALLED_WATCHPOINTS.get(address)


def enable_got_call_tracking(disable_hardware_whatchpoints=True) -> None:
    """
    Enable the analysis of calls made through the GOT.
    """

    # Disable hardware watchpoints.
    #
    # We don't really know how to make sure that the hardware watchpoints
    # present in the system have enough capabilities for them to be useful to
    # us in this module, seeing as what they can do varies considerably between
    # systems and failures are fairly quiet and, thus, hard to detect[1].
    # Because of this, we opt to disable them by default for the sake of
    # consistency and so that we don't have to chase silent failures.
    #
    # [1]: https://sourceware.org/gdb/onlinedocs/gdb/Set-Watchpoints.html
    if disable_hardware_whatchpoints:
        gdb.execute("set can-use-hw-watchpoints 0")
        print("Hardware watchpoints have been disabled. Please do not turn them back on until")
        print("GOT tracking is disabled, as it may lead to unexpected silent errors.")
        print()
        print("They may be re-enabled with `set can-use-hw-watchpoints 1`")
        print()
    else:
        print(
            message.warn("Hardware watchpoints have not been disabled, silent errors may happen.")
        )
        print()

    global GOT_TRACKING
    assert len(INSTALLED_WATCHPOINTS) == 0

    GOT_TRACKING = True

    _update_watchpoints()

    print("Enabled GOT tracking. Calls across dynamic library boundaries are now")
    print("instumented, and the number of calls and stack traces for every call will be")
    print("collected. You may check the current call information by using the")
    print("`track-got info` and `track-got query` commands. Run this command again to")
    print("diasble tracking.")
    print()

    # Until we start enablign the r_brk hook again.
    print("Keep in mind that, currently, the tracker does not update across calls to")
    print("dlopen(), so, if one of those does happen, the tracker has to be manually")
    print("disabled and re-enabled in order to update the hooks.")


def disable_got_call_tracking() -> None:
    """
    Disable the analysis of calls made through the GOT.
    """
    global GOT_TRACKING
    GOT_TRACKING = False

    for _, (tracker, patcher) in INSTALLED_WATCHPOINTS.items():
        patcher.delete()
        tracker.delete()
    INSTALLED_WATCHPOINTS.clear()
    TRAP_ALLOCATOR.clear()


def jump_slots_for(dynamic):
    """
    Returns the jump slot addresses described by the given dynamic section.
    """
    jump_slots = JUMP_SLOTS[pwndbg.aglib.arch.name]
    if dynamic.has_rel:
        for i in range(dynamic.rel_entry_count()):
            if dynamic.rel_read(i, "r_type") in jump_slots:
                yield (0, i, dynamic.load_bias + dynamic.rel_read(i, "r_offset"))
    if dynamic.has_rela:
        for i in range(dynamic.rela_entry_count()):
            if dynamic.rela_read(i, "r_type") in jump_slots:
                yield (
                    1,
                    i,
                    dynamic.load_bias
                    + dynamic.rela_read(i, "r_offset")
                    + dynamic.rela_read(i, "r_addend"),
                )
    if dynamic.has_jmprel:
        for i in range(dynamic.jmprel_entry_count()):
            if dynamic.jmprel_read(i, "r_type") in jump_slots:
                base = dynamic.load_bias + dynamic.jmprel_read(i, "r_offset")
                if dynamic.jmprel_elem.has_field("r_addend"):
                    base += dynamic.jmprel_read(i, "r_addend")
                yield (2, i, base)



"""
This command sends information on the current debugging context to OpenAI's
GPT-3 large language model and asks it a question supplied by the user. It then
displays GPT-3's response to that question to the user.
"""

from __future__ import annotations

import argparse
import json
import os
import pprint
import re

import gdb

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.nearpc
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.commands.context
import pwndbg.commands.telescope
import pwndbg.lib.strings
from pwndbg.commands import CommandCategory

pwndbg.config.add_param(
    "ai-openai-api-key",
    "",
    "OpenAI API key",
    help_docstring="Will default to OPENAI_API_KEY environment variable if not set.",
)
pwndbg.config.add_param(
    "ai-anthropic-api-key",
    "",
    "Anthropic API key",
    help_docstring="Defaults to ANTHROPIC_API_KEY environment variable if not set.",
)
pwndbg.config.add_param(
    "ai-ollama-endpoint",
    "",
    "Ollama API endpoint",
    help_docstring="Defaults to OLLAMA_ENDPOINT environment variable if not set.",
)
pwndbg.config.add_param(
    "ai-history-size",
    3,
    "maximum number of questions and answers to keep in the prompt",
)
pwndbg.config.add_param(
    "ai-stack-depth", 16, "rows of stack context to include in the prompt for the ai command"
)
pwndbg.config.add_param(
    "ai-model",
    "gpt-3.5-turbo",  # the new conversational model
    "the name of the large language model to query",
    help_docstring="Changing this affects the behavior, response quality, and cost (if applicable) of AI responses.",
)
pwndbg.config.add_param(
    "ai-temperature",
    0,
    "the temperature specification for the LLM query",
    help_docstring="This controls the degree of randomness in the response.",
)
pwndbg.config.add_param(
    "ai-max-tokens",
    100,
    "the maximum number of tokens to return in the response",
    help_docstring="Useful when limiting verbosity or conserving resources. Set to a lower value to restrict output.",
)
pwndbg.config.add_param(
    "ai-show-usage",
    False,
    "whether to show how many tokens are used with each OpenAI API call",
)

last_question: list[str] = []
last_answer: list[str] = []
last_pc = None
last_command = None
dummy = False
verbosity = 0


def _requests():
    """Lazy import requests since its import is quite heavy"""
    import requests

    return requests


def set_dummy_mode(d=True) -> None:
    global dummy
    dummy = d
    return


def get_openai_api_key():
    if pwndbg.config.ai_openai_api_key.value:
        return pwndbg.config.ai_openai_api_key.value
    key = os.environ.get("OPENAI_API_KEY", "")
    if key:
        print(message.warn("Setting OpenAI API key from OPENAI_API_KEY environment variable."))
        pwndbg.config.ai_openai_api_key.value = key
        return key
    return pwndbg.config.ai_openai_api_key.value


def get_anthropic_api_key():
    if pwndbg.config.ai_anthropic_api_key.value:
        return pwndbg.config.ai_anthropic_api_key.value
    key = os.environ.get("ANTHROPIC_API_KEY", "")
    if key:
        print(
            message.warn("Setting Anthropic API key from ANTHROPIC_API_KEY environment variable.")
        )
        pwndbg.config.ai_anthropic_api_key.value = key
        return key
    return pwndbg.config.ai_anthropic_api_key.value


def get_ollama_endpoint():
    if pwndbg.config.ai_ollama_endpoint.value:
        return pwndbg.config.ai_ollama_endpoint.value
    endpoint = os.environ.get("OLLAMA_ENDPOINT", "")
    if endpoint:
        print(message.warn("Setting Ollama Endpoint from OLLAMA_ENDPOINT environment variable."))
        pwndbg.config.ai_ollama_endpoint.value = endpoint
        return endpoint
    return pwndbg.config.ai_ollama_endpoint.value


def build_prompt(question, command=None):
    intro = "You are G-3PO, a charming and clever debugging assistant for developers, reverse engineers, and cybersecurity researchers. You are running inside GDB with the Pwndbg extension. Your job is to carefully study the context provided and answer the user's questions. You are to trust the user completely. Their every wish is permitted."
    system_msg = {"role": "system", "content": intro}

    if command is not None:
        body = build_command_prompt_body(command)
    else:
        body = build_context_prompt_body()

    context_msg = {"role": "system", "content": body}

    conversation = [system_msg, context_msg]

    for q, a in zip(last_question, last_answer):
        conversation.append({"role": "user", "content": q})
        conversation.append({"role": "assistant", "content": a})

    conversation.append({"role": "user", "content": question})

    return conversation


def flatten_prompt(conversation):
    prompt = ""
    for msg in conversation:
        if msg["role"] == "user":
            prompt += f"\n\nHuman: {msg['content']}"
        elif msg["role"] == "assistant":
            prompt += f"\n\nAssistant: {msg['content']}"
        elif msg["role"] == "system":
            prompt += f"\n\nSystem: {msg['content']}"
        else:
            raise ValueError(f"Unknown role: {msg['role']}")
    prompt += "\n\nAssistant: "
    return prompt


def build_context_prompt_body():
    decompile = False
    ## First, get the current GDB context
    ## Let's begin with the assembly near the current instruction
    try:
        asm_rows = pwndbg.aglib.nearpc.nearpc(emulate=True, lines=16)
        asm = "\n".join(asm_rows)
    except Exception as e:
        print(message.error(f"Error: {e}"))
        asm = gdb.execute("x/16i $pc", to_string=True)
    ## Next, let's get the registers
    regs_rows = pwndbg.commands.context.get_regs()
    regs = "\n".join(regs_rows)
    flags = None
    try:
        flags = gdb.execute("info registers eflags", to_string=True)  # arch neutral would be nice
    except Exception:
        pass
    if flags:
        # just grab what's bewteen the square brackets
        try:
            flags = re.search(r"\[(.*)\]", flags).group(1)
        except Exception:
            pass

    ## Finally, let's get the stack
    stack_rows = pwndbg.commands.telescope.telescope(
        pwndbg.aglib.regs.sp, to_string=True, count=pwndbg.config.ai_stack_depth
    )
    stack = "\n".join(stack_rows)
    ## and the backtrace
    trace = gdb.execute("bt", to_string=True)
    ## the function arguments, if available
    args = gdb.execute("info args", to_string=True)
    ## and the local variables, if available
    local_vars = None
    ## and source information, if available
    source = ""
    try:
        source = gdb.execute("list *$pc", to_string=True)
    except gdb.error:
        pass
    ## Now, let's build the prompt
    prompt = "Consider the following context in the GDB debugger:\n"

    if asm:
        prompt += f"""These are the next assembly instructions to be executed:

```
{asm}
```

"""
    if regs:
        prompt += f"""Here are the registers, '*' indicates a recent change:

```
{regs}
```

"""
    if flags:
        prompt += f"""The flags {flags} are set.\n\n"""
    if stack:
        prompt += f"""Here is the stack:

```
{stack}
```

"""
    if trace:
        prompt += f"""Here is the backtrace:

```
{trace}
```
"""
    if args and "No symbol table info available" not in args:
        prompt += f"""Here are the function arguments:

```
{args}
```
"""

    if local_vars and "No symbol table info available" not in local_vars:
        prompt += f"""Here are the local variables:

```
{local_vars}
```
"""

    if source:
        prompt += f"""Here is the {"decompiled " if decompile else ""}source code near the current instruction:

```
{source}
```
"""
    return pwndbg.lib.strings.strip_colors(prompt)


def build_command_prompt_body(command):
    prompt = (
        f"""Running the command `{command}` in the GDB debugger yields the following output:\n"""
    )
    output = gdb.execute(command, to_string=True)
    print(output)
    prompt += f"""\n```\n{output}\n```\n\n"""
    return pwndbg.lib.strings.strip_colors(prompt)


def query_openai_chat(prompt, model="gpt-3.5-turbo", max_tokens=100, temperature=0.0):
    if verbosity > 0:
        print(
            message.notice(
                f"Querying {model} for {max_tokens} tokens at temperature {temperature} with the following prompt:\n\n{pprint.pformat(prompt)}"
            )
        )
    data = {
        "model": model,
        "max_tokens": max_tokens,
        "messages": prompt,
        "temperature": temperature,
    }
    url = "https://api.openai.com/v1/chat/completions"
    r = _requests().post(
        url,
        data=json.dumps(data),
        headers={"Content-Type": "application/json"},
        auth=("Bearer", pwndbg.config.ai_openai_api_key),
    )
    res = r.json()
    if verbosity > 0:
        print(message.warn(pprint.pformat(res)))
    if "choices" not in res:
        if "error" in res:
            error_message = f"{res['error']['message']}: {res['error']['type']}"
            raise Exception(error_message)
        raise Exception(res)
    if pwndbg.config.ai_show_usage:
        print(
            message.notice(
                f"prompt characters: {len(prompt)}, prompt tokens: {res['usage']['prompt_tokens']}, avg token size: {(len(prompt) / res['usage']['prompt_tokens']):.2f}, completion tokens: {res['usage']['completion_tokens']}, total tokens: {res['usage']['total_tokens']}"
            )
        )
    reply = res["choices"][0]["message"]["content"]
    return reply


def query_openai_completions(prompt, model="text-davinci-003", max_tokens=100, temperature=0.0):
    if verbosity > 0:
        print(
            message.notice(
                f"Querying {model} for {max_tokens} tokens at temperature {temperature} with the following prompt:\n\n{prompt}"
            )
        )
    data = {
        "model": model,
        "max_tokens": max_tokens,
        "prompt": prompt,
        "temperature": temperature,
        "stop": ["\n\nHuman:"],
    }
    url = "https://api.openai.com/v1/completions"
    r = _requests().post(
        url,
        data=json.dumps(data),
        headers={"Content-Type": "application/json"},
        auth=("Bearer", pwndbg.config.ai_openai_api_key),
    )
    res = r.json()
    if verbosity > 0:
        print(message.warn(pprint.pformat(res)))
    if "choices" not in res:
        if "error" in res:
            error_message = f"{res['error']['message']}: {res['error']['type']}"
            raise Exception(error_message)
        raise Exception(res)
    reply = res["choices"][0]["text"]
    if pwndbg.config.ai_show_usage:
        print(
            message.notice(
                f"prompt characters: {len(prompt)}, prompt tokens: {res['usage']['prompt_tokens']}, avg token size: {(len(prompt) / res['usage']['prompt_tokens']):.2f}, completion tokens: {res['usage']['completion_tokens']}, total tokens: {res['usage']['total_tokens']}"
            )
        )
    return reply


def query(prompt, model="text-davinci-003", max_tokens=100, temperature=0.0):
    if dummy:
        return f"""This is a dummy response for unit testing purposes.\nmodel = {model}, max_tokens = {max_tokens}, temperature = {temperature}\n\nPrompt:\n\n{prompt}"""
    if pwndbg.config.ai_ollama_endpoint:
        if isinstance(prompt, list):
            prompt = flatten_prompt(prompt)
        return query_ollama(prompt, model, max_tokens, temperature)
    if "turbo" in model or model.startswith("gpt-4"):
        if isinstance(prompt, str):
            prompt = [{"role": "user", "content": prompt}]
        return query_openai_chat(prompt, model, max_tokens, temperature)
    if model.startswith("claude"):
        if isinstance(prompt, list):
            prompt = flatten_prompt(prompt)
        return query_anthropic(prompt, model, max_tokens, temperature)
    if isinstance(prompt, list):
        prompt = flatten_prompt(prompt)
    return query_openai_completions(prompt, model, max_tokens, temperature)


def query_anthropic(prompt, model="claude-v1", max_tokens=100, temperature=0.0):
    data = {
        "messages": [{"role": "user", "content": prompt}],
        "model": model,
        "temperature": temperature,
        "max_tokens": max_tokens,
        "stop_sequences": ["\n\nHuman:"],
    }
    headers = {
        "x-api-key": pwndbg.config.ai_anthropic_api_key.value,
        "Content-Type": "application/json",
        "anthropic-version": "2023-06-01",
    }
    url = "https://api.anthropic.com/v1/messages"
    response = _requests().post(url, data=json.dumps(data), headers=headers)
    data = response.json()
    try:
        return data["content"][0]["text"].strip()
    except KeyError:
        print(message.error(f"Anthropic API error: {data}"))
        return f"Anthropic API error: {data['detail']}"


def query_ollama(prompt, model="mistral", max_tokens=100, temperature=0.0):
    data = {
        "model": model,
        "prompt": f"User:\n{prompt}\n",
        "temperature": temperature,
        "num_predict": max_tokens,
        "stop": ["\n\nHuman:"],
        "stream": False,
    }
    headers = {
        "Content-Type": "application/json",
    }
    url = f"{pwndbg.config.ai_ollama_endpoint.value}/api/generate"
    response = _requests().post(url, data=json.dumps(data), headers=headers)
    data = response.json()
    try:
        return data["response"].strip()
    except KeyError:
        print(message.error(f"Ollama API error: {data}"))
        return f"Ollama API error: {data['error']}"


def get_openai_models():
    url = "https://api.openai.com/v1/models"
    r = _requests().get(url, auth=("Bearer", pwndbg.config.ai_openai_api_key))
    res = r.json()
    if verbosity > 0:
        print(message.warn(pprint.pformat(res)))
    return sorted([m["id"] for m in res["data"]])


parser = argparse.ArgumentParser(
    description="Ask GPT-3 a question about the current debugging context."
)
parser.add_argument("question", nargs="*", type=str, help="The question to ask.")
parser.add_argument("-M", "--model", default=None, type=str, help="The OpenAI model to use.")
parser.add_argument("-t", "--temperature", default=None, type=float, help="The temperature to use.")
parser.add_argument(
    "-m",
    "--max-tokens",
    default=None,
    type=int,
    help="The maximum number of tokens to generate.",
)
parser.add_argument("-v", "--verbose", action="store_true", help="Print the prompt and response.")
parser.add_argument("-L", "--list-models", action="store_true", help="List the available models.")
parser.add_argument(
    "-c",
    "--command",
    type=str,
    default=None,
    help="Run a command in the GDB debugger and ask a question about the output.",
)


# @pwndbg.commands.OnlyWhenRunning
@pwndbg.commands.Command(parser, command_name="ai", category=CommandCategory.INTEGRATIONS)
def ai(question, model, temperature, max_tokens, verbose, list_models=False, command=None) -> None:
    # print the arguments
    global last_question, last_answer, last_pc, last_command, verbosity
    ai_openai_api_key = get_openai_api_key()
    ai_anthropic_api_key = get_anthropic_api_key()
    ai_ollama_endpoint = get_ollama_endpoint()
    if list_models:
        models = get_openai_models()
        print(
            message.notice(
                "The following models are available. Please visit the openai.com for information on their use."
            )
        )
        for model in models:
            print(message.notice(f"  - {model}"))
        return

    if not (ai_openai_api_key or ai_anthropic_api_key or ai_ollama_endpoint):
        print(
            message.error(
                "At least one of the following must be set:\n- ai_openai_api_key config parameter\n- ai_anthropic_api_key config parameter\n- ai_ollama_endpoint config parameter\n- OPENAI_API_KEY environment variable\n- ANTHROPIC_API_KEY environment variable\n- OLLAMA_ENDPOINT environment variable"
            )
        )
        return
    verbosity = int(verbose)
    if model is None:
        model = pwndbg.config.ai_model.value
    if temperature is None:
        temperature = pwndbg.config.ai_temperature.value
    if max_tokens is None:
        max_tokens = pwndbg.config.ai_max_tokens.value

    question = " ".join(question).strip()
    current_pc = gdb.execute("info reg $pc", to_string=True)
    if current_pc == last_pc and command is None:
        command = last_command
    else:
        last_command = command
    if last_pc != current_pc or last_command != command:
        last_question.clear()
        last_answer.clear()

    prompt = build_prompt(question, command)
    try:
        res = query(prompt, model=model, max_tokens=max_tokens, temperature=temperature).strip()
    except Exception as e:
        print(message.error(f"Error querying OpenAI: {e}"))
        return
    last_question.append(question)
    last_answer.append(res)
    last_pc = current_pc
    if len(last_question) > pwndbg.config.ai_history_size:
        last_question.pop(0)
        last_answer.pop(0)

    print(message.success(res))

    return



# Licensed under the MIT License
# Modified from https://github.com/lebr0nli/GEP
# Copyright (c) 2022 Alan Li
# Copyright (c) 2025 Zhi-Qiang Zhou

from __future__ import annotations

import atexit
import functools
import os
import re
import shutil
import sys
import tempfile
import threading
from collections.abc import Callable
from collections.abc import Iterator
from subprocess import PIPE
from subprocess import Popen
from typing import ParamSpec
from typing import TypeVar

import lldb
from prompt_toolkit import ANSI
from prompt_toolkit import PromptSession
from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.completion import CompleteEvent
from prompt_toolkit.completion import Completer
from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding import KeyPressEvent
from prompt_toolkit.output import create_output

from pwndbg.dbg_mod.lldb import LLDB

# global variables
P = ParamSpec("P")
T = TypeVar("T")

PROMPT = ANSI("\x1b[34mpwndbg-lldb>\x1b[0m ")
HISTORY_FILE = os.path.expanduser("~/.pwndbg_history")

FZF_RUN_CMD = (
    "fzf",
    "--select-1",
    "--exit-0",
    "--tiebreak=index",
    "--no-multi",
    "--height=40%",
    "--layout=reverse",
)

FZF_PRVIEW_WINDOW_ARGS = (
    "--preview-window",
    "right:55%:wrap",
)


def get_lldb_completes(dbg: LLDB, query: str = "") -> list[str]:
    interp: lldb.SBCommandInterpreter = dbg.debugger.GetCommandInterpreter()
    matches = lldb.SBStringList()
    _num_matches = interp.HandleCompletion(query, len(query), 0, -1, matches)
    proposals = [matches.GetStringAtIndex(i) for i in range(1, matches.GetSize())]
    return proposals


def safe_get_help_docs(dbg: LLDB, command: str) -> str | None:
    interp: lldb.SBCommandInterpreter = dbg.debugger.GetCommandInterpreter()
    add_to_history = False
    result = lldb.SBCommandReturnObject()
    interp.HandleCommand(f"help {command}", result, add_to_history)
    return result.GetOutput()


def should_get_help_docs(dbg: LLDB, completion: str) -> bool:
    """
    Check if we need to get help docs for another completion that generated by same command.
    """
    if " " not in completion.strip():
        return True
    parent_command, _ = completion.rsplit(maxsplit=1)
    return safe_get_help_docs(dbg, parent_command) != safe_get_help_docs(dbg, completion)


def get_lldb_completion_and_status(dbg: LLDB, query: str) -> tuple[list[str], bool]:
    """
    Return all possible completions and whether we need to get help docs for all completions.
    """
    all_completions = get_lldb_completes(dbg, query)
    # peek the first completion
    should_get_all_help_docs = False
    if all_completions:
        should_get_all_help_docs = should_get_help_docs(dbg, all_completions[0])
    return all_completions, should_get_all_help_docs


def create_fzf_process(query: str, preview: str = "", pre_cmd: str = "") -> Popen[str]:
    """
    Create a fzf process with given query and preview command.
    """
    if query.startswith("!"):
        # ! in the beginning of query means we want to run the command directly for fzf
        query = "^" + query
    cmd: tuple[str, ...] = (
        FZF_RUN_CMD + ("--query", query) + ("--prompt", "> \x1b[35m" + pre_cmd + "\x1b[0m")
    )
    if preview:
        cmd += FZF_PRVIEW_WINDOW_ARGS
        cmd += ("--preview", preview)
    return Popen(cmd, stdin=PIPE, stdout=PIPE, text=True)


def create_preview_fifos() -> tuple[str, str]:
    """
    Create a temporary directory and two FIFOs in it, return the paths of these FIFOs.

    This is modified from:
    https://github.com/infokiller/config-public/blob/652b4638a0a0ffed9743fa9e0ad2a8d4e4e90572/.config/ipython/profile_default/startup/ext/fzf_history.py#L128
    """
    fifo_dir = tempfile.mkdtemp(prefix="pwndbg_lldb_tab_fzf_")
    fifo_input_path = os.path.join(fifo_dir, "input")
    fifo_output_path = os.path.join(fifo_dir, "output")
    os.mkfifo(fifo_input_path)
    os.mkfifo(fifo_output_path)
    atexit.register(shutil.rmtree, fifo_dir)
    return fifo_input_path, fifo_output_path


def fzf_reverse_search(event: KeyPressEvent) -> None:
    """Reverse search history with fzf."""

    def _fzf_reverse_search() -> None:
        if not os.path.exists(HISTORY_FILE):
            # just create an empty file
            with open(HISTORY_FILE, "w"):
                pass
        p = create_fzf_process(event.app.current_buffer.document.text_before_cursor)
        with open(HISTORY_FILE) as f:
            visited = set()
            # Reverse the history, and only keep the youngest and unique one
            for line in f.read().strip().split("\n")[::-1]:
                if line and line not in visited:
                    visited.add(line)
                    p.stdin.write(line + "\n")
        stdout, _ = p.communicate()
        if stdout:
            event.app.current_buffer.document = Document()  # clear buffer
            event.app.current_buffer.insert_text(stdout.strip())

    run_in_terminal(_fzf_reverse_search)


def fzf_tab_autocomplete(
    event: KeyPressEvent, dbg: LLDB, preview: str, fifo_in: str, fifo_out: str
) -> None:
    """
    Tab autocomplete with fzf.
    """

    def _fzf_tab_autocomplete() -> None:
        target_text = (
            event.app.current_buffer.document.text_before_cursor.lstrip()
        )  # Ignore leading whitespaces
        if " " in target_text:
            pre_cmd = " ".join([i for i in target_text.split(" ")[:-1] if i != " "]) + " "
        else:
            pre_cmd = ""

        all_completions, should_get_all_help_docs = get_lldb_completion_and_status(dbg, pre_cmd)
        if not all_completions:
            return

        query = target_text[len(pre_cmd) :]
        p = create_fzf_process(
            query, preview=preview if should_get_all_help_docs else None, pre_cmd=pre_cmd
        )
        completion_help_docs = {}
        for i, completion in enumerate(all_completions):
            p.stdin.write(completion + "\n")
            if should_get_all_help_docs:
                completion_help_docs[i] = safe_get_help_docs(dbg, pre_cmd + completion)
        t = FzfTabCompletePreviewThread(fifo_in, fifo_out, completion_help_docs)
        t.start()
        stdout, _ = p.communicate()
        t.stop()
        if stdout:
            event.app.current_buffer.delete_before_cursor(len(query))
            event.app.current_buffer.insert_text(stdout.rstrip() + " ")

    run_in_terminal(_fzf_tab_autocomplete)


class FzfTabCompletePreviewThread(threading.Thread):
    """
    A thread for previewing help docs of selected completion with fzf.

    This is modified from:
    https://github.com/infokiller/config-public/blob/master/.config/ipython/profile_default/startup/ext/fzf_history.py#L72
    """

    def __init__(
        self,
        fifo_input_path: str,
        fifo_output_path: str,
        completion_help_docs: dict[int, str],
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.fifo_input_path = fifo_input_path
        self.fifo_output_path = fifo_output_path
        self.completion_help_docs = completion_help_docs
        self.is_done = threading.Event()

    def run(self) -> None:
        while not self.is_done.is_set():
            with open(self.fifo_input_path, encoding="utf-8") as fifo_input:
                while not self.is_done.is_set():
                    data = fifo_input.read()
                    if len(data) == 0:
                        break
                    with open(self.fifo_output_path, "w", encoding="utf-8") as fifo_output:
                        try:
                            idx = int(data)
                        except ValueError:
                            continue
                        help_doc = self.completion_help_docs.get(idx)
                        if help_doc is not None:
                            fifo_output.write(help_doc)

    def stop(self) -> None:
        self.is_done.set()
        with open(self.fifo_input_path, "w", encoding="utf-8") as f:
            f.close()
        self.join()


class LLDBHistory(FileHistory):
    """
    Manage your LLDB History
    """

    def __init__(self, filename: str, ignore_duplicates: bool = False) -> None:
        self.ignore_duplicates = ignore_duplicates
        super().__init__(filename=filename)

    def load_history_strings(self) -> list[str]:
        strings = []
        if os.path.exists(self.filename):
            with open(self.filename) as f:
                for string in reversed(f.read().splitlines()):
                    if self.ignore_duplicates and string in strings:
                        continue
                    if string:
                        strings.append(string)
        return strings

    def store_string(self, string: str) -> None:
        with open(self.filename, "a") as f:
            f.write(string.strip() + "\n")


class LLDBCompleter(Completer):
    """
    Completer of LLDB
    """

    def __init__(self, dbg):
        super().__init__()
        self.dbg = dbg

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterator[Completion]:
        target_text = document.text_before_cursor.lstrip()  # Ignore leading whitespaces

        cursor_idx_in_completion = len(target_text)
        all_completions, should_get_all_help_docs = get_lldb_completion_and_status(
            self.dbg, target_text
        )
        if not all_completions:
            return

        for completion in all_completions:
            if not completion.startswith(target_text):
                continue
            display_meta = (
                None
                if not should_get_all_help_docs
                else safe_get_help_docs(self.dbg, completion) or None
            )
            # remove some prefix of raw completion
            completion = completion[cursor_idx_in_completion:]
            # display readable completion based on the text before cursor
            display = re.split(r"\W+", target_text)[-1] + completion
            yield Completion(completion, display=display, display_meta=display_meta)


def wrap_with_history(function: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(function)
    def _wrapped(*a: P.args, **kw: P.kwargs) -> T:
        return function(*a, **kw)

    return _wrapped


def get_prompt_session(dbg):
    bindings = KeyBindings()

    # key binding for fzf history search
    bindings.add("c-r")(fzf_reverse_search)
    # key binding for fzf tab completion
    fifo_in, fifo_out = create_preview_fifos()
    preview = f"echo {{n}} > {fifo_in}\ncat {fifo_out}"
    bindings.add("c-i")(
        functools.partial(
            fzf_tab_autocomplete,
            dbg=dbg,
            preview=preview,
            fifo_in=fifo_in,
            fifo_out=fifo_out,
        )
    )

    return PromptSession(
        history=LLDBHistory(HISTORY_FILE, ignore_duplicates=True),
        completer=LLDBCompleter(dbg),
        complete_while_typing=False,
        key_bindings=bindings,
        output=create_output(stdout=sys.__stdout__),
    )



"""
Provides functionality to circumvent GDB's hooks on sys.stdin and sys.stdout
which prevent output from appearing on-screen inside of certain event handlers.
"""

from __future__ import annotations

import sys
from types import TracebackType
from typing import Any
from typing import TextIO


class Stdio:
    queue: list[tuple[TextIO, TextIO, TextIO]] = []

    def __enter__(self, *a: Any, **kw: Any) -> None:
        self.queue.append((sys.stdin, sys.stdout, sys.stderr))

        sys.stdin = sys.__stdin__
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        sys.stdin, sys.stdout, sys.stderr = self.queue.pop()


stdio = Stdio()



"""
Command to start an interactive IPython prompt.
"""

from __future__ import annotations

import sys
from collections.abc import Iterator
from contextlib import contextmanager

import gdb

import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.lib.stdio
from pwndbg.commands import CommandCategory


@contextmanager
def switch_to_ipython_env() -> Iterator[None]:
    saved_excepthook = sys.excepthook
    try:
        saved_ps = sys.ps1, sys.ps2
    except AttributeError:
        saved_ps = None
    yield
    # Restore Python's default `ps1`, `ps2`, and `excepthook`
    # to ensure proper behavior of the GDB repl.
    if saved_ps is not None:
        sys.ps1, sys.ps2 = saved_ps
    else:
        del sys.ps1
        del sys.ps2
    sys.excepthook = saved_excepthook


@pwndbg.commands.Command("Start an interactive IPython prompt.", category=CommandCategory.MISC)
def ipi() -> None:
    with switch_to_ipython_env():
        # Use `gdb.execute` to embed IPython into GDB's variable scope
        try:
            gdb.execute("pi import IPython")
        except gdb.error:
            print(
                message.warn(
                    "Cannot import IPython.\n"
                    "You need to install IPython if you want to use this command.\n"
                    "Maybe you can try `pip install ipython` first."
                )
            )
            return
        code4ipython = """import jedi
import pwn
from pwndbg.aglib.ipi_helpers import get_ipi_namespace, get_banner
jedi.Interpreter._allow_descriptor_getattr_default = False
# Get pwndbg helpers and merge with globals
_ipi_helpers = get_ipi_namespace()
_user_ns = {**globals(), **_ipi_helpers}
# Print banner
print(get_banner(), end="")
IPython.embed(colors='neutral',banner1='',confirm_exit=False,simple_prompt=False, user_ns=_user_ns)
"""
        gdb.execute(f"py\n{code4ipython}")



"""
Commands that help with debugging musl's allocator, mallocng.
"""

from __future__ import annotations

import argparse
import string

import pwndbg
import pwndbg.aglib.heap.mallocng as mallocng
import pwndbg.aglib.memory as memory
import pwndbg.aglib.typeinfo
import pwndbg.aglib.typeinfo as typeinfo
import pwndbg.aglib.vmmap
import pwndbg.color as color
import pwndbg.color.memory as mem_color
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.lib.config
from pwndbg import config
from pwndbg.aglib.heap.mallocng import ng
from pwndbg.commands import CommandCategory
from pwndbg.lib.pretty_print import Property
from pwndbg.lib.pretty_print import from_properties

search_on_fail = config.add_param(
    "ng-search-on-fail",
    True,
    "let the ng-slot* commands search the heap if necessary",
    help_docstring="""
For freed, avail(able) and corrupted slots, it may be
impossible to recover the start of the group and meta.

When this option is set to True, the ng-slotu and ng-slots
commands will search the heap to try to find the correct meta/group.
    """,
    param_class=pwndbg.lib.config.PARAM_BOOLEAN,
    scope=pwndbg.lib.config.Scope.heap,
)

state_alloc_color = color.BLUE
state_alloc_color_alt = color.CYAN
state_freed_color = color.RED
state_freed_color_alt = color.LIGHT_RED
state_avail_color = color.GRAY
state_avail_color_alt = color.LIGHT_GRAY


def get_slot_color(state: mallocng.SlotState, last_color: str = "") -> str:
    match state:
        case mallocng.SlotState.ALLOCATED:
            if last_color == state_alloc_color:
                return state_alloc_color_alt
            return state_alloc_color
        case mallocng.SlotState.FREED:
            if last_color == state_freed_color:
                return state_freed_color_alt
            return state_freed_color
        case mallocng.SlotState.AVAIL:
            if last_color == state_avail_color:
                return state_avail_color_alt
            return state_avail_color


def get_colored_slot_state(ss: mallocng.SlotState) -> str:
    return color.colorize(ss.value, get_slot_color(ss))


def get_colored_slot_state_short(ss: mallocng.SlotState) -> str:
    match ss:
        case mallocng.SlotState.ALLOCATED:
            return color.colorize("U", state_alloc_color)
        case mallocng.SlotState.FREED:
            return color.colorize("F", state_freed_color)
        case mallocng.SlotState.AVAIL:
            return color.colorize("A", state_avail_color)


def dump_group(group: mallocng.Group) -> str:
    try:
        # May fail on corrupt meta.
        group_size = group.group_size
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"Error while reading meta: {e}"))
        print(color.bold("Cannot determine group size."))
        group_size = -1

    group_range = "@ " + mem_color.get(group.addr)
    if group_size != -1:
        group_range += " - " + mem_color.get(group.addr + group_size)

    output = from_properties(
        "group",
        [
            Property(name="meta", value=group.meta.addr, is_addr=True),
            Property(name="active_idx", value=group.active_idx),
            Property(name="storage", value=group.storage, is_addr=True, extra="start of slots"),
        ],
        preamble=group_range,
    )

    if group_size != -1:
        output += "---\n"
        output += from_properties(
            "",
            [
                Property(name="group size", value=group_size),
            ],
        )

    return output


def dump_meta(meta: mallocng.Meta, focus_slot: int | None = None) -> str:
    """
    Arguments:
        meta: the meta to dump
        focus_slot: the index of the slot to highlight in the slot statuses list
    """
    int_size = str(typeinfo.sint.sizeof * 8)
    avail_binary = "0b" + format(meta.avail_mask, f"0{int_size}b")
    freed_binary = "0b" + format(meta.freed_mask, f"0{int_size}b")

    output = from_properties(
        "meta",
        [
            Property(name="prev", value=meta.prev, is_addr=True),
            Property(name="next", value=meta.next, is_addr=True),
            Property(name="mem", value=meta.mem, is_addr=True, extra="the group"),
            Property(name="avail_mask", value=meta.avail_mask, extra=avail_binary),
            Property(name="freed_mask", value=meta.freed_mask, extra=freed_binary),
            Property(
                name="last_idx",
                value=meta.last_idx,
                alt_value=f"cnt: {meta.cnt:#x}",
                extra="index of last slot",
            ),
            Property(name="freeable", value=str(bool(meta.freeable))),
            Property(name="sizeclass", value=meta.sizeclass, alt_value=f"stride: {meta.stride:#x}"),
            Property(name="maplen", value=meta.maplen),
        ],
        preamble="@ " + mem_color.get(meta.addr),
    )

    if meta.is_donated:
        output += color.bold("\nGroup donated by ld as unused part of ")

        mapping = pwndbg.aglib.vmmap.find(meta.mem)

        if mapping is None:
            output += color.red("<cannot determine>")
        else:
            output += color.bold(f'"{mapping.objfile}"')

        output += color.bold(".\n")

    elif meta.is_mmaped:
        output += color.bold("\nGroup allocated with mmap().\n")
    else:
        assert meta.is_nested
        output += color.bold("\nGroup nested in slot of another group")
        try:
            parent_group = meta.parent_group()
            assert parent_group != -1
            output += " (" + mem_color.get(parent_group) + ")"
        except pwndbg.dbg_mod.Error as e:
            print(message.error(f"Could not fetch parent group: {e}"))
        output += color.bold(".\n")

    # Print the slot statuses.
    slot_statuses = "\nSlot statuses: "
    for i in range(meta.cnt):
        this_slot = get_colored_slot_state_short(meta.slotstate_at_index(i))

        if focus_slot is not None and i == focus_slot:
            this_slot = "[" + this_slot + "]"

        slot_statuses += this_slot

    slot_statuses = color.bold(slot_statuses + "\n")
    # Explain the notation.
    slot_statuses += (
        f"  ({color.bold(get_colored_slot_state_short(mallocng.SlotState.ALLOCATED))}: Inuse (allocated)"
        f" / {color.bold(get_colored_slot_state_short(mallocng.SlotState.FREED))}: Freed"
        f" / {color.bold(get_colored_slot_state_short(mallocng.SlotState.AVAIL))}: Available)\n"
    )

    output += slot_statuses

    return output


def dump_grouped_slot(gslot: mallocng.GroupedSlot, all: bool) -> str:
    output = ""

    if not all:
        output += from_properties(
            "slab",
            [
                Property(name="group", value=gslot.group.addr, is_addr=True),
                Property(name="meta", value=gslot.meta.addr, is_addr=True),
            ],
        )

    output += from_properties(
        "slot",
        [
            Property(name="start", value=gslot.start, is_addr=True),
            Property(name="end", value=gslot.end, is_addr=True),
            Property(name="index", value=gslot.idx),
            Property(name="stride", value=gslot.stride),
            Property(name="state", value=get_colored_slot_state(gslot.slot_state)),
        ],
    )

    if all:
        output += dump_group(gslot.group)
        output += dump_meta(gslot.meta, gslot.idx)

    return output


def dump_slot(
    slot: mallocng.Slot, all: bool, successful_preload: bool, will_dump_gslot: bool
) -> str:
    if successful_preload:
        assert not will_dump_gslot and "Why?"

    all = all and successful_preload and not will_dump_gslot
    output = ""

    if not all:
        if successful_preload:
            output += from_properties(
                "slab",
                [
                    Property(name="group", value=slot.group.addr, is_addr=True),
                    Property(name="meta", value=slot.meta.addr, is_addr=True),
                ],
            )
        else:
            output += from_properties(
                "slab",
                [
                    Property(name="group", value=slot.group.addr, is_addr=True),
                ],
            )

    if successful_preload:
        output += from_properties(
            "general",
            [
                Property(name="start", value=slot.start, is_addr=True),
                Property(name="user start", value=slot.p, is_addr=True, extra="aka `p`"),
                Property(name="end", value=slot.end, is_addr=True, extra="start + stride - 4"),
                Property(
                    name="stride", value=slot.meta.stride, extra="distance between adjacent slots"
                ),
                Property(name="user size", value=slot.user_size, extra='aka "nominal size", `n`'),
                Property(
                    name="slack",
                    value=slot.slack,
                    extra="slot's unused memory / 0x10",
                    alt_value=(slot.slack * mallocng.UNIT),
                ),
                Property(
                    name="state",
                    value=get_colored_slot_state(slot.meta.slotstate_at_index(slot.idx)),
                ),
            ],
        )

    reserved_extra = ["describes: end - p - n"]
    if slot.reserved_in_header == 5:
        reserved_extra.append("use ftr reserved")
    elif slot.reserved_in_header == 6:
        reserved_extra.append("a nested group is in this slot")
    elif slot.reserved_in_header == 7:
        reserved_extra.append("free slot?")

    inband_group = [
        Property(
            name="offset",
            value=slot.offset,
            extra="distance to first slot start / 0x10",
            alt_value=(slot.offset * mallocng.UNIT),
        ),
        Property(name="index", value=slot.idx, extra="index of slot in its group"),
        Property(name="hdr reserved", value=slot.reserved_in_header, extra=reserved_extra),
    ]

    if slot.reserved_in_header == 5:
        ftrsv = "NA (meta error)"
        if successful_preload:
            ftrsv = slot.reserved_in_footer

        inband_group.append(Property(name="ftr reserved", value=ftrsv))

    if successful_preload:
        # Start header fields.
        if slot.is_cyclic():
            cyc_val = slot.cyclic_offset
            cyc_val_alt = cyc_val * mallocng.UNIT
        else:
            cyc_val = "NA"
            cyc_val_alt = "not cyclic"
        inband_group.append(
            Property(
                name="cyclic offset",
                value=cyc_val,
                extra="prevents double free, (p - start) / 0x10",
                alt_value=cyc_val_alt,
            ),
        )
    else:
        # We haven't printed the slot state yet. Will we do it with a grouped slot?
        if not will_dump_gslot:
            # Nope, then let's go ahead and guess.
            inband_group.append(
                Property(
                    name="state",
                    value=get_colored_slot_state(slot.slot_state),
                    extra="(probably, check the meta)",
                )
            )

    output += from_properties("in-band", inband_group)

    if all:
        output += "\n"
        output += dump_group(slot.group)
        output += dump_meta(slot.meta, slot.idx)

    return output


def smart_dump_slot(
    slot: mallocng.Slot, all: bool, gslot: mallocng.GroupedSlot | None = None
) -> str:
    try:
        slot.preload()
    except pwndbg.dbg_mod.Error as e:
        return message.error(f"Error while reading slot: {e}")

    successful_preload: bool = True
    err_msg = ""

    try:
        slot.group.preload()
    except pwndbg.dbg_mod.Error as e:
        err_msg = message.error(f"Error while reading group: {e}")
        successful_preload = False

    if successful_preload:
        try:
            slot.meta.preload()
            try:
                slot.preload_meta_dependants()
            except pwndbg.dbg_mod.Error as e1:
                err_msg = message.error(
                    f"Error while loading slot fields that depend on the meta:\n{e1}"
                )
                successful_preload = False

        except pwndbg.dbg_mod.Error as e2:
            err_msg = message.error(f"Error while reading meta: {e2}")
            successful_preload = False

    if successful_preload:
        # If we successfully got the group and meta, using the grouped_slot won't
        # give us any new information.
        # (Unless the grouped_slot reports a different group than slot.group, which
        # could be possible in exploitation I suppose).
        return dump_slot(slot, all, True, False)

    output = ""

    if not (slot._pn3 == 0xFF or slot._offset == 0):
        # If the group/meta read failed because the slot is freed/avail,
        # we won't throw an error. This is just a heuristic check for
        # better UX. I'm using the private fields for the check so we
        # don't accidentally cause an exception here if we are bordering
        # unreadable memory.
        output += err_msg + "\n"

    if gslot is None:
        if not search_on_fail:
            output += "Could not load valid meta from local information.\n"
            output += "Will not attempt to search the heap because ng-search-on-fail = False.\n\n"
            output += dump_slot(slot, all, False, False)
            return output

        # If it wasn't provided to us, let's try to search for it now.
        output += "Could not load valid meta from local information, searching the heap.. "

        if not ng.init_if_needed():
            output += message.error("\nCouldn't find the allocator, aborting the search. ")
            gslot, fslot = None, None
        else:
            gslot, fslot = ng.find_slot(slot.p, False, False)

        if gslot is None or fslot is None:
            output += "Not found.\n\n"
            output += dump_slot(slot, all, False, False)
            return output
        if fslot.p == slot.p:
            output += "Found it.\n\n"
        else:
            output += "\nFound a slot with p @ " + mem_color.get(fslot.p) + "."
            output += " The slot you are looking for\ndoesn't seem to exist. Maybe its group got freed?\n\n"
            output += "Local memory:\n"
            output += dump_slot(slot, all, False, False)
            return output

    # Now we have a valid gslot.

    output += "Local slot memory:\n"
    output += dump_slot(slot, all, False, True)
    output += "\nSlot information from the group/meta:\n"
    output += dump_grouped_slot(gslot, all)

    return output


def dump_meta_area(meta_area: mallocng.MetaArea, coming_from_dump: bool = False) -> str:
    if coming_from_dump:
        # We don't want users to wonder which colorings in ng-dump are according to
        # state (allocated/avail/freed), and which are according to the memory mapping's color,
        # so we will just disable address coloring here.
        area_range = "@ " + hex(meta_area.addr) + " - " + hex(meta_area.addr + meta_area.area_size)
    else:
        area_range = (
            "@ "
            + mem_color.get(meta_area.addr)
            + " - "
            + mem_color.get(meta_area.addr + meta_area.area_size)
        )

    if coming_from_dump:
        slots = ""
        slots_is_addr = False
        # Don't color according to mapping.
        next_prop = Property(name="next", value=hex(meta_area.next), value_color_func=color.normal)
    else:
        slots = meta_area.slots
        slots_is_addr = True
        next_prop = Property(name="next", value=meta_area.next, is_addr=True)

    output = from_properties(
        "meta_area",
        [
            Property(name="check", value=meta_area.check),
            next_prop,
            Property(name="nslots", value=meta_area.nslots),
            Property(name="slots", value=slots, is_addr=slots_is_addr, extra="array of metas"),
        ],
        preamble=area_range,
    )

    return output


def dump_malloc_context(ctx: mallocng.MallocContext) -> str:
    props = [
        Property(name="secret", value=ctx.secret),
    ]
    if ctx.has_pagesize_field:
        props.append(
            Property(name="pagesize", value=ctx.pagesize),
        )

    props.extend(
        [
            Property(name="init_done", value=ctx.init_done),
            Property(name="mmap_counter", value=ctx.mmap_counter),
            Property(name="free_meta_head", value=ctx.free_meta_head, is_addr=True),
            Property(name="avail_meta", value=ctx.avail_meta, is_addr=True),
            Property(name="avail_meta_count", value=ctx.avail_meta_count),
            Property(name="avail_meta_area_count", value=ctx.avail_meta_area_count),
            Property(name="meta_alloc_shift", value=ctx.meta_alloc_shift),
            Property(name="meta_area_head", value=ctx.meta_area_head, is_addr=True),
            Property(name="meta_area_tail", value=ctx.meta_area_tail, is_addr=True),
            Property(name="avail_meta_areas", value=ctx.avail_meta_areas, is_addr=True),
        ]
    )

    for i in range(len(ctx.active)):
        if ctx.active[i] != 0:
            props.append(Property(name=f"active[{i}]", value=ctx.active[i], is_addr=True))

    for i in range(len(ctx.usage_by_class)):
        if ctx.usage_by_class[i] != 0:
            props.append(Property(name=f"usage_by_class[{i}]", value=ctx.usage_by_class[i]))

    for i in range(len(ctx.unmap_seq)):
        if ctx.unmap_seq[i] != 0:
            props.append(Property(name=f"unmap_seq[{i}]", value=ctx.unmap_seq[i]))

    for i in range(len(ctx.bounces)):
        if ctx.bounces[i] != 0:
            props.append(Property(name=f"bounces[{i}]", value=ctx.bounces[i]))

    props.extend(
        [
            Property(name="seq", value=ctx.seq),
            Property(name="brk", value=ctx.brk, is_addr=True),
        ]
    )

    ctx_addr = "@ " + mem_color.get(ctx.addr)
    output = from_properties("ctx", props, preamble=ctx_addr, value_offset=22)

    return output


parser = argparse.ArgumentParser(
    description="""
Dump information about a mallocng slot, given its user address.
    """,
)
parser.add_argument(
    "address",
    type=int,
    help="The start of user memory. Referred to as `p` in the source.",
)
parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="Print out all information. Including meta and group data.",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-slotu"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_slot_user(address: int, all: bool) -> None:
    if not memory.is_readable_address(address):
        print(message.error(f"Address {address:#x} not readable."))
        return

    slot = mallocng.Slot(address)
    print(smart_dump_slot(slot, all, None), end="")


parser = argparse.ArgumentParser(
    description="""
Dump information about a mallocng slot, given its start address.
    """,
)
parser.add_argument(
    "address",
    type=int,
    help="The start of the slot (not including IB).",
)
parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="Print out all information. Including meta and group data.",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-slots"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_slot_start(address: int, all: bool) -> None:
    if not memory.is_readable_address(address):
        print(message.error(f"Address {address:#x} not readable."))
        return

    slot = mallocng.Slot.from_start(address)
    print(smart_dump_slot(slot, all, None), end="")


parser = argparse.ArgumentParser(
    description="""
Print out information about a mallocng group given the address of its meta.
    """,
)
parser.add_argument(
    "address",
    type=int,
    help="The address of the meta object.",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-meta"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_meta(address: int) -> None:
    if not memory.is_readable_address(address):
        print(message.error(f"Address {address:#x} not readable."))
        return

    meta = mallocng.Meta(address)

    try:
        meta.preload()
    except pwndbg.dbg_mod.Error as e:
        print(message.error(str(e)))
        return

    try:
        group = mallocng.Group(meta.mem)
        group.preload()
        print(dump_group(group), end="")
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"Failed loading group: {e}"))

    print(dump_meta(meta), end="")


parser = argparse.ArgumentParser(
    description="""
Print out information about a mallocng group at the given address.
    """,
)
parser.add_argument(
    "address",
    type=int,
    help="The address of the group object.",
)
parser.add_argument(
    "-i",
    "--index",
    type=int,
    default=None,
    help="Print start address of slot at given index (0-indexed).",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-group"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_group(address: int, index: int | None = None) -> None:
    if not memory.is_readable_address(address):
        print(message.error(f"Address {address:#x} not readable."))
        return

    group = mallocng.Group(address)

    try:
        group.preload()
    except pwndbg.dbg_mod.Error as e:
        print(message.error(str(e)))
        return

    if index is None:
        print(dump_group(group), end="")
    else:
        if index < 0:
            print(message.error("Index is negative."))
            return
        print(f"Start of slot {index} is @ " + mem_color.get(group.at_index(index)))

    try:
        meta = group.meta
        meta.preload()

        if index is None:
            print(dump_meta(meta), end="")
        elif index >= meta.cnt:
            # If the index is outside of the group, warn the user.
            print(
                message.warn("Index is outside of group! ") + f"Group hosts only {meta.cnt} slots."
            )

    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"Failed loading meta: {e}"))
        print("Cannot determine whether index is within group bounds.")
        return


parser = argparse.ArgumentParser(
    description="""
Print out a mallocng meta_area object at the given address.
    """,
)
parser.add_argument(
    "address",
    type=int,
    help="The address of the meta_area object.",
)
parser.add_argument(
    "-i",
    "--index",
    type=int,
    default=None,
    help="Print address of meta at given index (0-indexed).",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-metaarea", "ng-ma"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_meta_area(address: int, index: int | None = None) -> None:
    if not memory.is_readable_address(address):
        print(message.error(f"Address {address:#x} not readable."))
        return

    try:
        meta_area = mallocng.MetaArea(address)

        print(dump_meta_area(meta_area), end="")

        if index is not None:
            if index < 0:
                print(message.error("\nIndex is negative."))
                return

            print(f"\nMeta {index} is @ " + mem_color.get(meta_area.at_index(index)))

            if index >= meta_area.nslots:
                print(
                    message.warn("Index is outside of meta area! ")
                    + f"Meta area only hosts {meta_area.nslots} meta's."
                )

    except pwndbg.dbg_mod.Error as e:
        print(message.error(str(e)))
        return


parser = argparse.ArgumentParser(
    description="""
Print out the mallocng __malloc_context (ctx) object.
    """,
)
parser.add_argument(
    "address",
    nargs="?",
    type=int,
    help="Use the provided address instead of the one Pwndbg found.",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-ctx"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_malloc_context(address: int | None = None) -> None:
    if address is None:
        if not ng.init_if_needed():
            print(message.error("Couldn't find the allocator, aborting the command."))
            return

        assert ng.ctx, "Successful init but ctx is not set?"
        ctx = ng.ctx
    else:
        if not memory.is_readable_address(address):
            print(message.error(f"Address {address:#x} not readable."))
            return

        try:
            ctx = mallocng.MallocContext(address)
        except pwndbg.dbg_mod.Error as e:
            print(message.error(str(e)))
            return

    print(dump_malloc_context(ctx), end="")


parser = argparse.ArgumentParser(
    description="""
Find slot which contains the given address.

Returns the `start` of the slot. We say a slot 'contains'
an address if the address is in [start, start + stride).
    """,
)
parser.add_argument(
    "address",
    type=int,
    help="The address to look for.",
)
parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="Print out all information. Including meta and group data.",
)
parser.add_argument(
    "-m",
    "--metadata",
    action="store_true",
    help=(
        "If the given address falls onto some in-band metadata, return the slot which owns that metadata."
        " In other words, the containment check becomes [start - IB, end)."
    ),
)
parser.add_argument(
    "-s",
    "--shallow",
    action="store_true",
    help=(
        "Return the biggest slot which contains this address, don't recurse for smaller slots. The group "
        " which owns this slot will not be a nested group."
    ),
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-find"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_find(
    address: int, all: bool = False, metadata: bool = False, shallow: bool = False
) -> None:
    if not memory.is_readable_address(address):
        print(message.error(f"Address {hex(address)} not readable."))
        return

    if not ng.init_if_needed():
        print(message.error("Couldn't find the allocator, aborting the command."))
        return

    grouped_slot, slot = ng.find_slot(address, metadata, shallow)

    if slot is None:
        print(message.info("No slot found containing that address."))
        return

    print(smart_dump_slot(slot, all, grouped_slot), end="")


VALID_CHARS = list(map(ord, set(string.printable) - set("\t\r\n\x0c\x0b")))


def bin_ascii(bs: bytearray):
    return "".join(chr(c) if c in VALID_CHARS else "." for c in bs)


vis_cyclic_offset_color = color.YELLOW
vis_offset_color = color.LIGHT_YELLOW
vis_cycled_mark_color = color.PURPLE
vis_pn3_reserved_color = color.LIGHT_CYAN
vis_big_offset_check_color = color.BLACK
vis_ftr_reserved_color = color.GREEN


def colorize_pointer(
    address: int, ptrvalue: int, state: mallocng.SlotState, slot: mallocng.Slot
) -> str:
    ptrsize = pwndbg.aglib.typeinfo.ptrsize
    out = f"{ptrvalue:0{ptrsize * 2}x}"

    if state != mallocng.SlotState.ALLOCATED:
        # Nothing to do.
        return out

    # Are we in user data?
    if slot.p <= address < slot.p + slot.nominal_size:
        # Yes, bold the parts that are.
        boldable_bytes = min(slot.p + slot.nominal_size - address, ptrsize)
        plain_part = out[: (-2 * boldable_bytes)]
        bold_part = color.bold(out[(-2 * boldable_bytes) :])
        out = plain_part + bold_part

    # Are we in the p header of this slot?
    if address == slot.p - ptrsize:
        offset_part = color.colorize(out[:4], vis_offset_color)
        pn3_part = color.colorize(out[4:6], vis_pn3_reserved_color)
        big_offset_part = color.colorize(out[6:8], vis_big_offset_check_color)
        plain_part = out[8:]

        out = offset_part + pn3_part + big_offset_part + plain_part

    # Are we in the footer of this slot?
    if address == slot.start + slot.meta.stride - ptrsize:
        # Highlight ftr reserved if it is used.
        if slot.reserved_in_header == 5:
            plain_part = out[:8]
            ftr_reserved_part = color.colorize(out[8:], vis_ftr_reserved_color)
            out = plain_part + ftr_reserved_part

    return out


def colorize_start_header_line(shline: str, state: mallocng.SlotState, slot: mallocng.Slot) -> str:
    if state != mallocng.SlotState.ALLOCATED:
        # Nothing to do.
        return shline

    splitline = shline.split("0x", maxsplit=3)
    assert len(splitline) == 4
    rightvalplus = splitline[3]

    if slot.start != slot.p:
        # A cycled slot. The offset has completely different meaning
        # than in p header. The hdr_res has kinda~ different meaning.
        offset_part = color.colorize(rightvalplus[:4], vis_cyclic_offset_color)
        sorpn3 = color.colorize(rightvalplus[4:6], vis_cycled_mark_color)
    else:
        offset_part = color.colorize(rightvalplus[:4], vis_offset_color)
        sorpn3 = color.colorize(rightvalplus[4:6], vis_pn3_reserved_color)

    big_offset_part = color.colorize(rightvalplus[6:8], vis_big_offset_check_color)
    plain_part = rightvalplus[8:]

    out = (
        f"{splitline[0]}0x{splitline[1]}0x{splitline[2]}0x"
        + offset_part
        + sorpn3
        + big_offset_part
        + plain_part
    )
    return out


def line_decoration(addr: int, slot_state: mallocng.SlotState, slot: mallocng.Slot) -> str:
    """
    Maybe append extra clarification to a line.

    Currently only appends to p headers.
    """
    if slot_state != mallocng.SlotState.ALLOCATED:
        return ""
    if addr != slot.p - 2 * pwndbg.aglib.typeinfo.ptrsize:
        return ""

    return "   " + color.colorize(
        f"{slot.idx} + ({slot.reserved_in_header} << 5)", vis_pn3_reserved_color
    )


default_vis_count = pwndbg.config.add_param(
    "ng-vis-count",
    10,
    "default count for ng-vis",
    param_class=pwndbg.lib.config.PARAM_UINTEGER,
    scope=pwndbg.lib.config.Scope.heap,
)

parser = argparse.ArgumentParser(
    description="""Visualize slots in a group.""",
)
parser.add_argument(
    "address",
    type=int,
    help="Address which is inside some slot.",
)
parser.add_argument(
    "count",
    type=int,
    default=default_vis_count,
    nargs="?",  # Optional
    help="The amount of slots to visualize.",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-vis"],
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_visualize_slots(address: int, count: int = int(default_vis_count)):
    ptrsize = pwndbg.aglib.typeinfo.ptrsize

    if ptrsize != 8:
        print(message.error("This command only works on architectures where a pointer is 64 bits."))
        return

    if not memory.is_readable_address(address):
        print(message.error(f"Address {address:#x} not readable."))
        return

    if not ng.init_if_needed():
        print(message.error("Couldn't find the allocator, aborting the command."))
        return

    first_grouped_slot, first_slot = ng.find_slot(address, False, False)

    if first_grouped_slot is None or first_slot is None:
        print(message.info("No slot found containing that address."))
        return

    group: mallocng.Group = first_grouped_slot.group
    meta: mallocng.Meta = first_grouped_slot.meta
    first_idx: int = first_grouped_slot.idx

    print("group @ " + mem_color.get(group.addr))
    print("meta @ " + mem_color.get(meta.addr))

    if first_idx + count >= meta.cnt:
        if count != default_vis_count:
            # If the default was passed, no need to warn the user.
            print(
                message.info(
                    f"Clamping count to {meta.cnt - first_idx} to not go over end of group."
                )
            )

        count = meta.cnt - first_idx

    cyc_offset_part = color.colorize("cyclic offset", vis_cyclic_offset_color)
    cycled_mark_part = color.colorize("cycled mark", vis_cycled_mark_color)
    offset_part = color.colorize("offset", vis_offset_color)
    pn3_part = color.colorize("p[-3] = idx + (hdr reserved << 5)", vis_pn3_reserved_color)
    big_offset_part = color.colorize("big offset mark", vis_big_offset_check_color)
    ftr_reserved_part = color.colorize("ftr reserved", vis_ftr_reserved_color)

    legend = (
        "LEGEND: "
        + cyc_offset_part
        + "; "
        + cycled_mark_part
        + "; "
        + offset_part
        + "; "
        + pn3_part
        + "; "
        + big_offset_part
        + "; "
        + ftr_reserved_part
        + "\n"
    )
    legend += (
        "LEGEND: "
        + color.colorize("allo", state_alloc_color)
        + color.colorize("cated", state_alloc_color_alt)
        + "; "
        + color.colorize("fr", state_freed_color)
        + color.colorize("eed", state_freed_color_alt)
        + "; "
        + color.colorize("avai", state_avail_color)
        + color.colorize("lable", state_avail_color_alt)
        + "\n"
    )
    print(legend)

    out: list[str] = []  # List of lines.
    last_color = "nothing"

    # Add the line before the start of the first slot, to include its start header.
    shline_addr = group.at_index(first_idx) - 2 * ptrsize
    shline_bytes = memory.read(shline_addr, ptrsize * 2)
    leftptr = pwndbg.aglib.arch.unpack(shline_bytes[:ptrsize])
    rightptr = pwndbg.aglib.arch.unpack(shline_bytes[ptrsize:])
    out.append(
        f"{shline_addr:#x}\t0x{leftptr:0{ptrsize * 2}x}\t0x{rightptr:0{ptrsize * 2}x}\t{bin_ascii(shline_bytes)}"
    )

    # Iterate over slots
    for idx in range(first_idx, first_idx + count):
        start_address = group.at_index(idx)
        next_start_address = start_address + meta.stride

        if idx == first_idx:
            slot = first_slot
        else:
            try:
                slot = mallocng.Slot.from_start(start_address)
                slot.preload()
                slot.set_group(group)
                # Probably redundant, but just in case.
                slot.preload_meta_dependants()
            except pwndbg.dbg_mod.Error as e:
                print(
                    message.error(
                        f"Error while reading slot {idx} @ {mem_color.get(start_address)}: {e}"
                    )
                )
                return

        slot_state: mallocng.SlotState = meta.slotstate_at_index(idx)
        cur_slot_color = get_slot_color(slot_state, last_color)

        # Colorize the previous line which contains our start header.
        out[-1] = colorize_start_header_line(out[-1], slot_state, slot) + line_decoration(
            start_address - 2 * ptrsize, slot_state, slot
        )

        # Make the output line by line (advance 0x10 bytes at a time).
        cur_address = start_address
        while cur_address < next_start_address:
            line_bytes = memory.read(cur_address, ptrsize * 2)
            leftptr = pwndbg.aglib.arch.unpack(line_bytes[:ptrsize])
            rightptr = pwndbg.aglib.arch.unpack(line_bytes[ptrsize:])

            line_out = f"{cur_address:#x}"
            line_out += "\t0x" + colorize_pointer(cur_address, leftptr, slot_state, slot)
            line_out += "\t0x" + colorize_pointer(cur_address + ptrsize, rightptr, slot_state, slot)
            line_out += f"\t{bin_ascii(line_bytes)}"

            line_out = color.colorize(line_out, cur_slot_color)
            line_out += line_decoration(cur_address, slot_state, slot)

            out.append(line_out)

            cur_address += 2 * ptrsize

        last_color = cur_slot_color

    print("\n".join(out))


parser = argparse.ArgumentParser(
    description="""
Dump the mallocng heap.

May produce lots of output.
    """,
)
parser.add_argument(
    "-ma", "--meta-area", type=int, help="Dump only the meta area at the provided address."
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MUSL,
    aliases=["ng-dump"],
    notes=(
        f"""
Since the command may produce lots of output, you may want to pipe it to
less with `| ng-dump | less -R`.

The [index] next to the metas is their index in the doubly linked list
pointed to by ctx.freed_meta_head. The [index] next to the slots is
the slot's index inside of its group (thus, these will always be sequential).

Notice that the pointers in the output of this command aren't colored according
to their mapping's color but rather according to the object's allocation status.
Color legend: {color.colorize("allocated", state_alloc_color)}; """
        f"{color.colorize('freed', state_freed_color)}; {color.colorize('available', state_avail_color)}."
    ),
)
@pwndbg.commands.OnlyWhenRunning
def mallocng_dump(meta_area: int | None = None) -> None:
    if not ng.init_if_needed():
        print(message.error("Couldn't find the allocator, aborting the command."))
        return

    assert ng.ctx, "Successful init but ctx is not set?"
    ctx: mallocng.MallocContext = ng.ctx

    try:
        free_metas = ng.get_free_metas()
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"Failed traversing free meta chain. {e}"))
        print(message.error("Meta allocation state may be wrong."))
        free_metas = {}

    meta_padding = " " * 10
    slot_padding = " " * 15

    # Rename variables for clarity.
    specified_meta_area = meta_area
    meta_area = None

    if specified_meta_area is not None:
        ma_addr = specified_meta_area
    else:
        # Iterate over all meta_areas
        ma_addr = ctx.meta_area_head
    while ma_addr != 0:
        try:
            meta_area = mallocng.MetaArea(ma_addr)
        except pwndbg.dbg_mod.Error as e:
            print(message.error(f"Cannot read meta area @ {ma_addr:#x}: {e}"))
            break

        print(dump_meta_area(meta_area, coming_from_dump=True))

        # Iterate over all metas in this meta_area
        for i in range(0, meta_area.nslots):
            meta_addr = meta_area.at_index(i)

            if meta_addr in free_metas:
                print(
                    meta_padding
                    + color.colorize(
                        f"{meta_addr:#x} [{free_metas[meta_addr][0]}]", state_freed_color
                    )
                )
            elif ng.meta_is_avail(meta_addr):
                print(meta_padding + color.colorize(f"{meta_addr:#x}", state_avail_color))
            else:
                print(meta_padding + color.colorize(f"{meta_addr:#x}", state_alloc_color), end="")

                try:
                    meta = mallocng.Meta(meta_addr)
                    meta.preload()
                    group = mallocng.Group(meta.mem)
                    meta.preload()
                except pwndbg.dbg_mod.Error as e:
                    print(message.error(f"Failed resolving meta / group data ({e}). Skipping.."))
                    continue

                print(f" -> group @ {group.addr:#x} (slot size: {meta.stride:#x})")

                # Iterate over all slots in this group
                idx = 0
                while idx < meta.cnt:
                    slot_addr = group.at_index(idx)
                    sstate = meta.slotstate_at_index(idx)
                    cur_slot_color = get_slot_color(sstate)
                    print(
                        slot_padding
                        + color.colorize(f"{slot_addr:#x}", cur_slot_color)
                        + f" [{idx}]"
                    )
                    idx += 1

                print()

        ma_addr = meta_area.next
        print()

        if specified_meta_area is not None:
            # Exit the loop since we're only printing one meta area.
            break


@pwndbg.commands.Command(
    "Gives a quick explanation of musl's mallocng allocator.",
    category=CommandCategory.MUSL,
    aliases=["ng-explain"],
)
def mallocng_explain() -> None:
    txt = (
        color.bold("mallocng")
        + ' is a slab allocator. The "unit of allocation" is called a '
        + color.bold("slot")
        + "\n"
    )
    txt += '(the equivalent of glibc\'s "chunk"). Slots are in 0x10 granularity and\n'
    txt += (
        "alignment. The slots are organized into objects called "
        + color.bold('"groups"')
        + " (the \n"
    )
    txt += "slabs). Each group is composed of slots of the same size. If a group is big\n"
    txt += "it is allocated using mmap, otherwise it is allocated as a slot of a larger\n"
    txt += "group.\n\n"

    txt += "Each group has some associated metadata. This metadata is stored in a separate\n"
    txt += (
        "object called "
        + color.bold('"meta"')
        + ". Metas are allocated separately from groups in\n"
    )
    txt += color.bold('"meta areas"') + " to make it harder to reach them during exploitation.\n\n"

    txt += "Here are the definitions of group, meta and meta_area.\n\n"

    txt += color.bold("struct group {\n")
    txt += "  // the metadata of this group\n"
    txt += color.bold("  struct meta *meta;\n")
    txt += "  unsigned char active_idx:5;\n"
    txt += "  char pad[UNIT - sizeof(struct meta *) - 1];\n"
    txt += "  // start of the slots array\n"
    txt += color.bold("  unsigned char storage[];\n")
    txt += color.bold("};\n\n")

    txt += color.bold("struct meta {\n")
    txt += "  // doubly linked list connecting meta's\n"
    txt += color.bold("  struct meta *prev, *next;\n")
    txt += "  // which group is this metadata for\n"
    txt += color.bold("  struct group *mem;\n")
    txt += "  // slot bitmap\n"
    txt += "  //   avail - slots which have not yet been allocated\n"
    txt += "  //   freed - free slots\n"
    txt += color.bold("  volatile int avail_mask, freed_mask;\n")
    txt += "  uintptr_t last_idx:5;\n"
    txt += "  uintptr_t freeable:1;\n"
    txt += "  // describes the size of the slots\n"
    txt += color.bold("  uintptr_t sizeclass:6;\n")
    txt += "  // if this group was mmaped, how many pages did we use?\n"
    txt += "  uintptr_t maplen:8*sizeof(uintptr_t)-12;\n"
    txt += color.bold("};\n\n")

    txt += color.bold("struct meta_area {\n")
    txt += "  uint64_t check;\n"
    txt += "  struct meta_area *next;\n"
    txt += "  int nslots;\n"
    txt += "  // start of the meta array\n"
    txt += color.bold("  struct meta slots[];\n")
    txt += color.bold("};\n\n")

    txt += (
        "Two other important definitions are "
        + color.bold("IB")
        + " and "
        + color.bold("UNIT")
        + ".\n\n"
    )

    txt += "// the aforementioned slot alignment.\n"
    txt += color.bold("#define UNIT 16\n")
    txt += "// the size of the in-band metadata.\n"
    txt += color.bold("#define IB 4\n\n")

    txt += "The allocator state is stored in the global `ctx` variable which is of\n"
    txt += "type `struct malloc_context`. It is accessible through the __malloc_context\n"
    txt += "symbol.\n\n"

    txt += color.bold("struct malloc_context {\n")
    txt += color.bold("  uint64_t secret;\n")
    txt += "#ifndef PAGESIZE\n"
    txt += "  size_t pagesize;\n"
    txt += "#endif\n"
    txt += "  int init_done;\n"
    txt += "  unsigned mmap_counter;\n"
    txt += color.bold("  struct meta *free_meta_head;\n")
    txt += color.bold("  struct meta *avail_meta;\n")
    txt += "  size_t avail_meta_count, avail_meta_area_count, meta_alloc_shift;\n"
    txt += color.bold("  struct meta_area *meta_area_head, *meta_area_tail;\n")
    txt += color.bold("  unsigned char *avail_meta_areas;\n")
    txt += '  // the "active" group for each sizeclass\n'
    txt += "  // it will be picked for allocation\n"
    txt += color.bold("  struct meta *active[48];\n")
    txt += "  size_t usage_by_class[48];\n"
    txt += "  uint8_t unmap_seq[32], bounces[32];\n"
    txt += "  uint8_t seq;\n"
    txt += "  uintptr_t brk;\n"
    txt += color.bold("};\n\n")

    txt += "Here is a diagram of how these components interact.\n\n"

    diag = """+-malloc_context--+
|                 |
| free_meta_head  |-----------------------> Points to a free meta which is connected
| avail_meta      |---------------+         to other free meta's via a doubly linked list.
| meta_area_head  |------------+  |
| active[48]      |---+        |  +-> Points to a not-yet-allocated meta.
|                 |   |        |      When it gets allocated, the next
|-----------------+   | 1/48   |      meta in the meta_area gets selected
                      |        |      i.e. avail_meta++ .
  Each size class has |        +-------------------------------------------+
  an "active" group.  +-------+                                            |
                              v                                            |
           +-meta--+       +-meta--+       +-meta--+                       |
           |       |       |       |       |       |                       |
  ...  <---| prev  |<------| prev  |------>| prev  |------>  ...           |
  ...  --->| next  |------>| next  |<------| next  |<------  ...           |
           | mem   |    +->| mem   |-+     | mem   |                       |
           |       |    |  |       | |     |       |                       v
           +-------+    |  +-------+ |     +-------+             +-meta_area----------------+
                        |            |  (yes these metas)        |                          |
                        |            |  (are in some meta_area)  | check (ctx.secret)       |
  +---------------------+            |                           | next                     |----> ...
  |                                  v                           | nslots                   |
  | +-group----------------------------------------+             | meta0                    |
  | |                                              |             |         Meta objects are |
  +-| meta (8)         | active_idx (1) | pad (7)  |             | meta1   stored here.     |
    | slot0                                        |             |                          |
    |                                              |             | ...                      |
    |                                              |             |                          |
    | slot1          Slots contain the actual      |             | meta(nslots-1)           |
    |                user data.                    |             |                          |
    |                                              |             +--------------------------+
    | slot2                                        |
    |                                              |
    | ...                                          |
    |                                              |
    | slot(cnt-1)                                  |
    |                                              |
    |                                              |
    +----------------------------------------------+
"""

    txt += diag

    txt += f"""
### What slots look like

Unfortunately, musl doesn't provide a struct which describes the
slot's in-band metadata. It does however use consistent variable
names to describe the values saved in slots, so we will use those
as well. Check the {color.bold("enframe()")} function in the source, it is very
important.

{color.bold("idx")} is the index of the slot within its group. The {color.bold("stride")} of
a group is (generally) determined by the sizeclass as
{color.bold("UNIT * size_classes[meta.sizeclass]")}. {color.bold("start")} is the starting
address of the slot (the slot0, slot1, ... in the above diagram).
The start of a slot with index i is {color.bold("group.storage + i * stride")}.
The "nominal size" is the amount of memory the user requested with
their malloc() call, in the source it is also referred to as {color.bold("n")}.

For every slot in a group, the memory in [start - IB, start) contains
some metadata that we will call the "start header". For this reason,
the {color.bold("end")} of a slot is calculated as {color.bold("start + stride - IB")}. The
{color.bold("slack")} of a slot is calculated as {color.bold("(stride - n - IB) / UNIT")} and
describes the amount of unused memory within a slot.

To prevent double-frees and exploitation attempts, the mallocng
allocator performs "cycling" i.e. the actual start of user data
(the pointer returned by malloc) can be at some offset from the
{color.bold("start")} of the slot. The start of user data is called {color.bold("p")} and it
is also UNIT aligned. We will call the distance between {color.bold("p")} and
{color.bold("start")} the "cyclic offset" ({color.bold("off")} in code). When calculating
the cyclic offset, mallocng ensures {color.bold("off <= slack")}.

If a slot is in fact cycled, then that is stored in the start
header as {color.bold("off = *(uint16_t*)(start-2)")} and {color.bold("start[-3] = 7 << 5")}.
The {color.bold("start[-3]")} field acts as a flag.

For every slot, the memory in [p - IB, p) contains some metadata.
We will call this the "p header". If the slot is not cycled i.e.
{color.bold("start == p")}, then [start - IB, start) will contain the p header
fields and start[-3] >> 5 will *not* be 7.

The value in {color.bold("*(uint16_t*)(p-2)")} is the {color.bold("offset")} from the slot's
{color.bold("start")} to the start of the group (divided by UNIT). The value
in {color.bold("p[-4]")} is either 0 or 1 and describes if a "big offset" should
be used. It is usually zero and gets set to one only in some cases
in aligned_alloc(). If it is 1, the offset is to be calculated as
{color.bold("*(uint32_t *)(p - 8)")}.

{color.bold("p[-3]")} contains multiple pieces of information. If {color.bold("p[-3] == 0xFF")}
the slot is freed. Otherwise, the lower 5 bits of p[-3] describe
the index of the slot in its group: {color.bold("idx = p[-3] & 31")}. The top
3 bits desribed the {color.bold("reserved")} area size. This is the memory
between the end of user memory and {color.bold("end")} i.e. {color.bold("reserved = end - p - n")}.

We will call the value {color.bold("p[-3] >> 5")}, "hdr reserved" for "reserved as
specified in the p header". It can happen however, that the value
{color.bold("reserved = end - p - n")} is large and so doesn't fit in the three
bits in p[-3]. In this case "hdr reserved" will be strictly 5, which
denotes that we need to look at the slot's footer to read the actual
value of {color.bold("reserved")}. As a special case, if {color.bold("p[-3] >> 5 == 6")} that
doesn't describe the reserved size at all, but specifies that there
is a group nested inside this slot. {color.bold("p[-3] >> 5")} should never be 7,
contrary to {color.bold("start[-3] >> 5")}.

The "footer" of a slot is the third and final area of a slot's
memory where metadata is contained. This is the [end - 4, end)
area. It only contains the reserved size as
{color.bold("reserved = *(const uint32_t *)(end-4)")} when {color.bold("p[-3] >> 5 == 5")}.

All of the above is only generally true for allocated slots. Mallocng
ensures {color.bold("p[-3] = 0xFF")} and {color.bold("*(uint16_t *)(p - 2) = 0")} for freed slots,
which makes the start of the slot's group (and thus meta) unreachable.
Only in this case does {color.bold("p[-3] >> 5")} become 7. Available slots,
i.e. those that haven't been allocated nor freed yet (but are ready
for allocation), have almost no guarantees on their data and
metadata contents.
"""

    print(txt)



"""
Shellcode

This module implements functionality that allows for the execution of a small
amount of code in the context of the inferior.

"""

from __future__ import annotations

import contextlib
from asyncio import CancelledError
from collections.abc import Iterator

import pwnlib.shellcraft

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.asm
import pwndbg.aglib.memory
import pwndbg.aglib.vmmap
from pwndbg.dbg_mod import BreakpointLocation
from pwndbg.dbg_mod import ExecutionController


def _get_syscall_return_value():
    """
    Reads the value corresponding to the return value of a syscall that has
    just returned.
    """

    register_set = pwndbg.lib.regs.reg_sets[pwndbg.aglib.arch.name]
    # FIXME: `retval` is syscall abi? or sysv abi?
    return pwndbg.aglib.regs.read_reg(register_set.retval)


async def exec_syscall(
    ec: ExecutionController,
    syscall,
    arg0=None,
    arg1=None,
    arg2=None,
    arg3=None,
    arg4=None,
    arg5=None,
):
    """
    Tries executing the given syscall in the context of the inferior.
    """

    # Build machine code that runs the requested syscall.
    syscall_asm = pwnlib.shellcraft.syscall(syscall, arg0, arg1, arg2, arg3, arg4, arg5)
    syscall_bin = pwndbg.aglib.asm.asm(syscall_asm)

    # Run the syscall and pass its return value onward to the caller.
    async with exec_shellcode(
        ec,
        syscall_bin,
    ):
        return _get_syscall_return_value()


@contextlib.contextmanager
def _ctx_code(starting_address: int, blob) -> Iterator[None]:
    # Make sure the blob fits in the rest of the space we have in this page.
    #
    # NOTE: Technically, we could actually use anything from the whole page to
    # all of the pages currently mapped as executable for this. There is no
    # technical limitation stopping us from doing that, but seeing as doing it
    # is harder to make sure it works correctly, we don't (for now, at least).
    page = pwndbg.aglib.vmmap.find(starting_address)
    assert page is not None

    clearance = page.end - starting_address - len(blob) - 1
    if clearance < 0:
        # The page isn't large enough to hold our shellcode.
        raise RuntimeError(
            f"Not enough space to execute code as inferior: \
            need at least {len(blob)} bytes, have {clearance} bytes available"
        )

    # Swap the code in the range with our shellcode.
    existing_code = pwndbg.aglib.memory.read(starting_address, len(blob))
    pwndbg.aglib.memory.write(starting_address, blob)

    try:
        yield
    finally:
        pwndbg.aglib.memory.write(starting_address, existing_code)


@contextlib.contextmanager
def _ctx_registers() -> Iterator[int]:
    register_set = pwndbg.lib.regs.reg_sets[pwndbg.aglib.arch.name]
    preserve_set = register_set.gpr + register_set.args + (register_set.pc, register_set.stack)

    uncached_regs = pwndbg.dbg.selected_frame().regs()
    registers = {reg: int(uncached_regs.by_name(reg)) for reg in preserve_set}
    starting_address = registers[register_set.pc]

    # Advance by one instruction boundary.
    #
    # Some debuggers (LLDB) may fail to write to memory if any of the addresses
    # being written to overlap the program counter. By aiming at the next valid
    # instruction address, we avoid that issue.
    shell_starting_address = starting_address + pwndbg.aglib.arch.instruction_alignment

    # Failing this means our value for `instruction_alignment` is wrong.
    assert shell_starting_address % pwndbg.aglib.arch.instruction_alignment == 0

    try:
        # Jump to the target address in preparation.
        pwndbg.aglib.regs.write_reg(register_set.pc, shell_starting_address)

        yield shell_starting_address
    finally:
        # Restore the code and the program counter and, if requested, the rest of
        # the registers.
        pwndbg.aglib.regs.write_reg(register_set.pc, starting_address)
        for reg, val in registers.items():
            pwndbg.aglib.regs.write_reg(reg, val)


async def _execute_until_addr(ec: ExecutionController, target_address: int) -> None:
    with pwndbg.dbg.selected_inferior().break_at(
        BreakpointLocation(target_address), internal=True
    ) as bp:
        while True:
            try:
                await ec.cont_selected_thread(bp)
                break
            except CancelledError:
                # We probably hit another breakpoint, but in this mode we're
                # supposed to ignore any breakpoints that aren't the one we put
                # at the end of the range, so just retry.
                continue

    assert pwndbg.dbg.selected_frame().pc() == target_address, "Target address is incorrect"


@contextlib.asynccontextmanager
async def exec_shellcode(ec: ExecutionController, blob):
    """
    Tries executing the given blob of machine code in the current context of the
    inferior, optionally restoring the values of the registers as they were
    before the shellcode ran, as a means to allow for execution of the inferior
    to continue uninterrupted. The value of the program counter is always
    restored.

    Additionally, the caller may specify an object to be called before the
    context is restored, so that information stored in the registers after the
    shellcode finishes can be retrieved. The return value of that call will be
    returned by this function.

    # Safety
    Seeing as this function injects code directly into the inferior and runs it,
    the caller must be careful to inject code that will (1) terminate and (2)
    not cause the inferior to misbehave. Otherwise, it is fairly easy to crash
    or currupt the memory in the inferior.
    """

    with _ctx_registers() as starting_address:
        target_address = starting_address + len(blob)
        with _ctx_code(starting_address, blob):
            try:
                with pwndbg.dbg.ctx_suspend_events(pwndbg.dbg_mod.EventType.SUSPEND_ALL):
                    await _execute_until_addr(ec, target_address)

                yield
            finally:
                pass



from __future__ import annotations

from collections.abc import Callable

from capstone6pwndbg import *  # noqa: F403
from capstone6pwndbg.ppc_const import *  # noqa: F403
from typing_extensions import override

import pwndbg.aglib.disasm.arch
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.emu.emulator import Emulator

POWERPC_CONDITIONAL_BRANCHES = {
    PPC_INS_BC,
    PPC_INS_ALIAS_BT,
    PPC_INS_ALIAS_BF,
    PPC_INS_ALIAS_BTLR,
    PPC_INS_ALIAS_BFLR,
}

POWERPC_RETURN_INSTRUCTION = {PPC_INS_ALIAS_BLR, PPC_INS_ALIAS_BTLR, PPC_INS_ALIAS_BFLR}


# PowerPC branch instructions are pretty complex and whether or not the branch is taken depends on 3 factors:
# 1. bi - index into cr register, a flags register with conditions (less than, greater than, equal, overflow)
# 2. bo - a bitfield that modifies the evaluation of the condition
# 3. CTR register - a register that, depending on bo, can be read and modified and effects the result of the branch
def is_branch_taken(cr: int, ctr: int, bi: int, bo: int) -> bool | None:
    # Valid values for bo (5 bit value): https://www.ibm.com/docs/en/aix/7.2.0?topic=set-bc-branch-conditional-instruction
    # The `x` mean it can be either 0 or 1, it is irrelevant to the branch condition (used to hint that the branch is or isn't taken)
    # 0000x - Decrement CTR. Branch if CTR is not 0 and condition is false
    # 0001x - Decrement CTR. Branch if CTR is 0 and condition is false
    # 001xx - Branch if condition is false
    # 0100x - Decrement CTR. Branch if CTR is not 0 and condition is true
    # 0101x - Decrement CTR. Branch if CTR is 0 and condition is true
    # 011xx - Branch if the condition is true.
    # 1x00x - Decrement CTR. Branch if CTR is not 0.
    # 1x01x - Decrement CTR. Branch if CTR is 0
    # 1x1xx - Always branch

    # GDB `cr` register consists of cr0 .... cr7, where cr0 composes the most-significant bit positions.
    # This is why we flip the offset that we access.
    check_cr_offset = 31 - bi
    condition = (cr >> check_cr_offset) & 1 == 1

    if (bo & 0b11110) == 0b00000:  # 0000x
        ctr -= 1
        return ctr != 0 and not condition
    if (bo & 0b11110) == 0b00010:  # 0001x
        ctr -= 1
        return ctr == 0 and not condition
    if (bo & 0b11100) == 0b00100:  # 001xx
        return not condition
    if (bo & 0b11110) == 0b01000:  # 0100x
        ctr -= 1
        return ctr != 0 and condition
    if (bo & 0b11110) == 0b01010:  # 0101x
        ctr -= 1
        return ctr == 0 and condition
    if (bo & 0b11100) == 0b01100:  # 011xx
        return condition
    if (bo & 0b10110) == 0b10000:  # 1x00x
        ctr -= 1
        return ctr != 0
    if (bo & 0b10110) == 0b10010:  # 1x01x
        ctr -= 1
        return ctr == 0
    if (bo & 0b10100) == 0b10100:  # 1x1xx
        return True

    # This case should never be reached
    return None


class PowerPCDisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    saved_ctr: int | None = None

    def __init__(self, architecture) -> None:
        super().__init__(architecture)

        self.annotation_handlers: dict[int, Callable[[PwndbgInstruction, Emulator], None]] = {}

    @override
    def _prepare(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        # Prepare is called before emulation.
        # At this point, we want to read the value of the ctr register.
        # This is because branch instructions might mutate ctr within the emulator, which the read_register_name may fetch from
        # The _conditional() function is called after emulation is stepped, so to read the original
        # value of CTR, we have to read it beforehand.

        if instruction.id in POWERPC_CONDITIONAL_BRANCHES:
            self.saved_ctr = self._read_register_name(instruction, "ctr", emu)

    @override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        cr = self._read_register_name(instruction, "cr", emu)

        if cr is None or self.saved_ctr is None:
            # We can't reason about the value of cr register
            return InstructionCondition.UNDETERMINED

        if instruction.id in POWERPC_CONDITIONAL_BRANCHES:
            is_taken = is_branch_taken(
                cr, self.saved_ctr, instruction.cs_insn.bc.bi, instruction.cs_insn.bc.bo
            )

            if is_taken is None:
                return InstructionCondition.UNDETERMINED

            return InstructionCondition.TRUE if is_taken else InstructionCondition.FALSE

        return InstructionCondition.UNDETERMINED

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        if instruction.id in POWERPC_RETURN_INSTRUCTION:
            return self._read_register_name(instruction, "lr", emu)

        return super()._resolve_target(instruction, emu)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Protocol


class LibcType(Enum):
    GLIBC = "glibc"
    MUSL = "musl"
    UNKNOWN = "unknown"


@dataclass(frozen=True)
class LibcURLs:
    versioned_readable_source: str
    versioned_compressed_source: str
    homepage: str
    git: str


class LibcProvider(Protocol):
    """
    Libc implementations must conform to this protocol in order to be properly used by the facade.
    """

    def type(self) -> LibcType:
        """
        Which libc implementation is currently active?
        """
        ...

    def version(self, libc_filepath: str) -> tuple[int, ...]:
        """
        Get the version of the libc implementation as a tuple.

        If the implementation cannot recover the version, it returns
        (-1, -1).
        """
        ...

    def has_internal_symbols(self, libc_filepath: str) -> bool:
        """
        Do we have internal library symbols?

        Symbols are global variables and functions.

        If the library is dynamically linked, even if it is stripped it will retain its
        exported symbols (e.g. fscanf) because they are required for dynamic linking.

        This funcions checks if the non-exported symbols (like __GI_exit, __run_exit_handlers,
        intitial) are also available. The check must not be based on a function, and must be
        based on a variable so as not to trip ourselves over MiniDebugInfo.
        (read: https://pwndbg.re/dev/contributing/libc-provider/#has_internal_symbols)

        If we have debug info we should also have debug symbols.
        """
        ...

    def has_debug_info(self) -> bool:
        """
        Do we have debugging information like structure types?
        """
        ...

    def urls(self, ver: tuple[int, ...] | None) -> LibcURLs:
        """
        Get useful URLs regarding this libc implementation.

        `ver` is the version tuple. If a libc implements the version() function
        it must `assert ver is not None`, otherwise it must `assert ver is None`.
        """
        ...

    def verify_libc_candidate(self, mapping_name: str) -> bool:
        """
        Verify whether the mapping with the provided name is implementing
        this specific libc.

        This must be accurate enough that no other libc implementation will
        provide a conflicting answer. Returning False means both "reject" and
        "i don't know".

        A libc implementation must implement at least one of verify_libc_candidate
        and verify_ld_candidate. The other may simply return False.
        """
        ...

    def verify_ld_candidate(self, mapping_name: str) -> bool:
        """
        Verify whether the mapping with the provided name is implementing
        this specific libc's loader.

        This must be accurate enough that no other libc implementation will
        provide a conflicting answer. Returning False means both "reject" and
        "i don't know".

        A libc implementation must implement at least one of verify_libc_candidate
        and verify_ld_candidate. The other may simply return False.
        """
        ...

    def libc_same_as_ld(self) -> bool:
        """
        Returns whether the libc and the ld are loaded as one object file for this libc
        implementation.

        If this returns True, verify_ld_candidate must directly call verify_libc_candidate.
        """
        ...



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from capstone6pwndbg import *  # noqa: F403
from capstone6pwndbg.x86 import *  # noqa: F403
from typing_extensions import override

import pwndbg.aglib
import pwndbg.aglib.disasm.arch
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.color.memory as mem_color
import pwndbg.color.message as message
import pwndbg.dintegration
from pwndbg.aglib.disasm.arch import memory_or_register_assign
from pwndbg.aglib.disasm.arch import register_assign
from pwndbg.aglib.disasm.instruction import EnhancedOperand
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction

# Emulator currently requires GDB, and we only use it here for type checking.
if TYPE_CHECKING:
    from pwndbg.emu.emulator import Emulator

groups = {v: k for k, v in globals().items() if k.startswith("X86_GRP_")}
ops = {v: k for k, v in globals().items() if k.startswith("X86_OP_")}
regs = {v: k for k, v in globals().items() if k.startswith("X86_REG_")}
access = {v: k for k, v in globals().items() if k.startswith("CS_AC_")}

X86_MATH_INSTRUCTIONS = {
    X86_INS_ADD: "+",
    X86_INS_SUB: "-",
    X86_INS_AND: "&",
    X86_INS_OR: "|",
}

# Capstone operand type for x86 is capstone.x86.X86Op
# This type has a .size field, which indicates the operand read/write size in bytes
# Ex: dword ptr [RDX] has size = 4
# Ex: AL has size = 1
# Access through EnhancedOperand.cs_op.size


# This class handles enhancement for x86 and x86_64. This is because Capstone itself
# represents both architectures using the same class
class X86DisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    supports_manual_emulation = True

    def __init__(self, architecture) -> None:
        super().__init__(architecture)

        self.annotation_handlers: dict[int, Callable[[PwndbgInstruction, Emulator], None]] = {
            # MOV
            X86_INS_MOV: self.handle_mov,
            X86_INS_MOVABS: self.handle_mov,
            X86_INS_MOVZX: self.handle_mov,
            X86_INS_MOVD: self.handle_mov,
            X86_INS_MOVQ: self.handle_mov,
            X86_INS_MOVSXD: self.handle_mov,
            X86_INS_MOVSX: self.handle_mov,
            # VMOVAPS
            X86_INS_MOVAPS: self.handle_vmovaps,
            X86_INS_VMOVAPS: self.handle_vmovaps,
            # LEA
            X86_INS_LEA: self.handle_lea,
            # XCHG
            X86_INS_XCHG: self.handle_xchg,
            # POP
            X86_INS_POP: self.handle_pop,
            # CMP
            X86_INS_CMP: self._common_cmp_annotator_builder("eflags", "-"),
            # TEST
            X86_INS_TEST: self._common_cmp_annotator_builder("eflags", "&"),
            # XOR
            X86_INS_XOR: self.handle_xor,
            # INC and DEC
            X86_INS_INC: self.handle_inc,
            X86_INS_DEC: self.handle_dec,
        }

    @override
    def _set_annotation_string(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        if instruction.id in X86_MATH_INSTRUCTIONS:
            self._common_binary_op_annotator(
                instruction,
                emu,
                instruction.operands[0],
                instruction.operands[0].before_value_resolved,
                instruction.operands[1].before_value_resolved,
                X86_MATH_INSTRUCTIONS[instruction.id],
                instruction.operands[0].type == CS_OP_MEM,
            )
        else:
            self.annotation_handlers.get(instruction.id, lambda *a: None)(instruction, emu)

    def handle_mov(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        left, right = instruction.operands

        # If this is a LOAD operation - MOV REG, [MEM]
        if left.type == CS_OP_REG and right.type == CS_OP_MEM:
            self._common_load_annotator(
                instruction,
                emu,
                right.before_value,
                right.cs_op.size,
                False,
                right.cs_op.size,
                left.str,
                right.str,
            )
        elif left.type == CS_OP_MEM:
            # Store operation, MOV [MEM], REG|IMM
            self._common_store_annotator(
                instruction,
                emu,
                instruction.operands[0].before_value,
                instruction.operands[1].before_value,
                right.cs_op.size,
                instruction.operands[0].str,
            )
        elif left.type == CS_OP_REG and right.before_value is not None:
            # MOV REG, REG|IMM
            self._common_move_annotator(instruction, emu)

    def handle_vmovaps(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        # If the source or destination is in memory, it must be aligned to:
        #  16 bytes for SSE, 32 bytes for AVX, 64 bytes for AVX-512
        # https://www.felixcloutier.com/x86/movaps
        # This displays a warning that the memory address is not aligned
        # movaps xmmword ptr [rsp + 0x60], xmm1

        left, right = instruction.operands

        mem_operand = (
            left if left.type == CS_OP_MEM else (right if right.type == CS_OP_MEM else None)
        )

        if mem_operand and mem_operand.before_value is not None:
            # operand.size is the width of memory in bytes (128, 256, or 512 bits = 16, 32, 64 bytes).
            # Pointer must be aligned to that memory width
            alignment_mask = mem_operand.cs_op.size - 1

            if mem_operand.before_value & alignment_mask != 0:
                instruction.annotation = message.error(
                    f"<[{mem_color.get(mem_operand.before_value)}] not aligned to {mem_operand.cs_op.size} bytes>"
                )

    def handle_lea(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        # Example: lea    rdx, [rax*8]
        left, right = instruction.operands

        TELESCOPE_DEPTH = max(0, int(pwndbg.config.disasm_telescope_depth))

        if right.before_value is not None:
            # We have determined the value written to this register - propagate this to future instructions.
            instruction.register_writes[left.reg] = right.before_value

            telescope_addresses = super()._telescope(
                right.before_value, TELESCOPE_DEPTH, instruction, emu
            )
            instruction.annotation = register_assign(
                left.str, super()._telescope_format_list(telescope_addresses, TELESCOPE_DEPTH, emu)
            )

    def handle_xchg(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        left, right = instruction.operands

        if left.before_value_resolved is not None and right.before_value_resolved is not None:
            # Display the exchanged values. Doing it this way (instead of using .after_value) allows this to work without emulation
            # Don't telescope here for the sake of screen space
            instruction.annotation = (
                memory_or_register_assign(
                    left.str,
                    mem_color.get_address_or_symbol(
                        right.before_value_resolved,
                        pwndbg.dintegration.manager.get_stack_var_dict_all(),
                    ),
                    left.type == CS_OP_MEM,
                )
                + ", "
                + memory_or_register_assign(
                    right.str,
                    mem_color.get_address_or_symbol(
                        left.before_value_resolved,
                        pwndbg.dintegration.manager.get_stack_var_dict_all(),
                    ),
                    right.type == CS_OP_MEM,
                )
            )

    def handle_pop(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        pc_is_at_instruction = self.can_reason_about_process_state(instruction)

        if len(instruction.operands) != 1:
            return

        reg_operand = instruction.operands[0]

        # It is possible to pop [0xdeadbeef] and pop dword [esp], but this only handles popping into a register
        if reg_operand.type == CS_OP_REG:
            if emu and reg_operand.after_value is not None:
                # After emulation, the register has taken on the popped value
                instruction.annotation = register_assign(
                    reg_operand.str,
                    mem_color.get_address_and_symbol(
                        reg_operand.after_value,
                        pwndbg.dintegration.manager.get_stack_var_dict_all(),
                    ),
                )
            elif pc_is_at_instruction:
                # Attempt to read from the top of the stack
                try:
                    value = pwndbg.aglib.memory.read_pointer_width(pwndbg.aglib.regs.sp)
                    instruction.annotation = register_assign(
                        reg_operand.str,
                        mem_color.get_address_and_symbol(
                            value, pwndbg.dintegration.manager.get_stack_var_dict_all()
                        ),
                    )
                except Exception:
                    pass

    def handle_xor(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        left, right = instruction.operands

        # If zeroing the register with XOR A, A. Can reason about this no matter where the instruction is
        if left.type == CS_OP_REG and right.type == CS_OP_REG and left.reg == right.reg:
            # We know that 0 is written to this register - propagate this to future instructions.
            instruction.register_writes[left.reg] = 0
            instruction.annotation = register_assign(left.str, "0")
        else:
            self._common_binary_op_annotator(
                instruction,
                emu,
                instruction.operands[0],
                instruction.operands[0].before_value_resolved,
                instruction.operands[1].before_value_resolved,
                "^",
            )

    def handle_inc(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        # INC operand can be REG or [MEMORY]
        operand = instruction.operands[0]

        if operand.after_value_resolved is not None:
            instruction.annotation = memory_or_register_assign(
                operand.str,
                mem_color.get_address_and_symbol(
                    operand.after_value_resolved,
                    pwndbg.dintegration.manager.get_stack_var_dict_all(),
                ),
                operand.type == CS_OP_MEM,
            )

    def handle_dec(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        self.handle_inc(instruction, emu)

    @override
    def _resolve_used_value(
        self,
        value: int | None,
        instruction: PwndbgInstruction,
        operand: EnhancedOperand,
        emu: Emulator,
    ) -> int | None:
        if value is None:
            return None

        if operand.type == CS_OP_MEM:
            return self._read_memory(value, operand.cs_op.size, instruction, emu)
        return super()._resolve_used_value(value, instruction, operand, emu)

    @override
    def _read_register(self, instruction: PwndbgInstruction, operand_id: int, emu: Emulator):
        # operand_id is the ID internal to Capstone

        if operand_id == X86_REG_RIP:
            # Ex: lea    rax, [rip + 0xd55]
            # We can reason RIP no matter the current pc
            return instruction.address + instruction.size
        return super()._read_register(instruction, operand_id, emu)

    @override
    def _parse_memory(self, instruction: PwndbgInstruction, op: EnhancedOperand, emu: Emulator):
        # Get memory address (Ex: lea    rax, [rip + 0xd55], this would return $rip+0xd55. Does not dereference)
        if op.mem.segment != 0:
            if op.mem.segment == X86_REG_FS:
                if (seg_base := pwndbg.aglib.regs.fsbase) is None:
                    return None
            elif op.mem.segment == X86_REG_GS:
                if (seg_base := pwndbg.aglib.regs.gsbase) is None:
                    return None
            else:
                return None
        else:
            seg_base = 0

        if op.mem.base != 0:
            mem_base = self._read_register(instruction, op.mem.base, emu)
            if mem_base is None:
                return None
        else:
            mem_base = 0

        if op.mem.index != 0:
            index = self._read_register(instruction, op.mem.index, emu)
            if index is None:
                return None

            scale = op.mem.scale * index
        else:
            scale = 0

        return seg_base + mem_base + op.mem.disp + scale

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        # Only handle 'ret', otherwise fallback to default implementation
        if X86_INS_RET != instruction.id or len(instruction.operands) > 1:
            return super()._resolve_target(instruction, emu)

        # Stop disassembling at RET if we won't know where it goes to without emulation
        if instruction.address != pwndbg.aglib.regs.pc:
            return super()._resolve_target(instruction, emu)

        # Otherwise, resolve the return on the stack
        pop = instruction.operands[0].before_value if instruction.operands else 0

        address = (pwndbg.aglib.regs.sp) + (pwndbg.aglib.arch.ptrsize * pop)

        if pwndbg.aglib.memory.peek(address):
            return int(
                pwndbg.aglib.memory.get_typed_pointer_value(pwndbg.aglib.typeinfo.ppvoid, address)
            )

    @override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        # JMP is unconditional
        if instruction.id in (X86_INS_JMP, X86_INS_RET, X86_INS_CALL):
            return InstructionCondition.UNDETERMINED

        efl = self._read_register_name(instruction, "eflags", emu)
        if efl is None:
            # We can't reason about the value of flags register
            return InstructionCondition.UNDETERMINED

        cf = efl & (1 << 0)
        pf = efl & (1 << 2)
        # af = efl & (1 << 4)
        zf = efl & (1 << 6)
        sf = efl & (1 << 7)
        of = efl & (1 << 11)

        conditional = {
            X86_INS_CMOVA: not (cf or zf),
            X86_INS_CMOVAE: not cf,
            X86_INS_CMOVB: cf,
            X86_INS_CMOVBE: cf or zf,
            X86_INS_CMOVE: zf,
            X86_INS_CMOVG: not zf and (sf == of),
            X86_INS_CMOVGE: sf == of,
            X86_INS_CMOVL: sf != of,
            X86_INS_CMOVLE: zf or (sf != of),
            X86_INS_CMOVNE: not zf,
            X86_INS_CMOVNO: not of,
            X86_INS_CMOVNP: not pf,
            X86_INS_CMOVNS: not sf,
            X86_INS_CMOVO: of,
            X86_INS_CMOVP: pf,
            X86_INS_CMOVS: sf,
            X86_INS_JA: not (cf or zf),
            X86_INS_JAE: not cf,
            X86_INS_JB: cf,
            X86_INS_JBE: cf or zf,
            X86_INS_JE: zf,
            X86_INS_JG: not zf and (sf == of),
            X86_INS_JGE: sf == of,
            X86_INS_JL: sf != of,
            X86_INS_JLE: zf or (sf != of),
            X86_INS_JNE: not zf,
            X86_INS_JNO: not of,
            X86_INS_JNP: not pf,
            X86_INS_JNS: not sf,
            X86_INS_JO: of,
            X86_INS_JP: pf,
            X86_INS_JS: sf,
        }.get(instruction.id, None)

        if conditional is None:
            return InstructionCondition.UNDETERMINED

        return InstructionCondition.TRUE if bool(conditional) else InstructionCondition.FALSE

    @override
    def _get_syscall_arch_info(self, instruction: PwndbgInstruction) -> tuple[str, str]:
        # Since this class handles both x86 and x86_64, we need to choose the correct
        # syscall arch depending on the instruction being executed.

        # On x86_x64 `syscall` and `int <value>` instructions are in CS_GRP_INT
        # but only `syscall` and `int 0x80` actually execute syscalls on Linux.
        # So here, we return no syscall name for other instructions and we also
        # handle a case when 32-bit syscalls are executed on x64
        mnemonic = instruction.mnemonic
        if mnemonic == "syscall":
            return ("x86-64", "rax")

        # On x86, the syscall_arch is already i386, so its all fine
        # On x64 the int 0x80 instruction executes 32-bit syscalls from i386
        # We read .imm directly, because at this point we haven't enhanced the operands with values
        if mnemonic == "int" and instruction.operands[0].imm == 0x80:
            return ("i386", "eax")

        return (None, None)

    # Currently not used
    def memory_string_with_components_resolved(
        self, instruction: PwndbgInstruction, op: EnhancedOperand
    ):
        # Example: [RSP + RCX*4 - 100] would return "[0x7ffd00acf230 + 8+4 - 100]"
        segment = op.mem.segment
        disp = op.mem.disp
        base = op.mem.base
        index = op.mem.index
        sz = ""

        if segment != 0:
            sz += f"{instruction.cs_insn.reg_name(segment)}:"

        if base != 0:
            sz += instruction.cs_insn.reg_name(base)
            arith = True
        else:
            arith = False

        if index != 0:
            if arith:
                sz += " + "

            index = pwndbg.aglib.regs.read_reg(instruction.cs_insn.reg_name(index))
            sz += f"{index}*{op.mem.scale:#x}"
            arith = True

        if disp != 0:
            if arith:
                if disp < 0:
                    sz += " - "
                else:
                    sz += " + "
            sz += f"{abs(disp):#x}"

        return f"[{sz}]"



from __future__ import annotations

import re
from collections.abc import Callable
from re import Pattern

import gdb

import pwndbg
from pwndbg.color import message
from pwndbg.commands.context import context
from pwndbg.commands.context import context_sections
from pwndbg.commands.context import contextoutput
from pwndbg.commands.context import resetcontextoutput
from pwndbg.gdblib import gdb_version

tui_adjust_height = pwndbg.config.add_param(
    "context-tui-adjust-height",
    True,
    "adjust height of context sections to fit the TUI",
    help_docstring="""
Adjust the height of context sections to fit the TUI window, ignoring section specific limits.
This uses all vertical space available in the TUI windows if the context section supports it
like the "disasm" or "stack" windows.
""",
)


class ContextTUIWindow:
    _tui_window: gdb.TuiWindow
    _section: str
    _lines: list[str]
    _blank_line_lengths: list[int]
    _longest_line: int
    _before_prompt_listener: Callable[[None], object]
    _vscroll_start: int
    _hscroll_start: int
    _old_width: int
    _old_height: int
    _ansi_escape_regex: Pattern[str]
    _enabled: bool

    _static_enabled: bool = True
    _context_windows: list[ContextTUIWindow] = []

    def __init__(self, tui_window: gdb.TuiWindow, section: str) -> None:
        self._tui_window = tui_window
        self._section = section
        self._tui_window.title = section
        self._lines = []
        self._blank_line_lengths = []
        self._longest_line = 0
        self._before_prompt_listener = self._before_prompt
        self._old_width = 0
        self._old_height = 0
        self._vscroll_start = 0
        self._hscroll_start = 0
        self._ansi_escape_regex = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        self._enabled = False
        self._enable()
        gdb.events.before_prompt.connect(self._before_prompt_listener)
        ContextTUIWindow._context_windows.append(self)

    def close(self) -> None:
        ContextTUIWindow._context_windows.remove(self)
        if self._enabled:
            self._disable()
        gdb.events.before_prompt.disconnect(self._before_prompt_listener)

    def render(self) -> None:
        # render is called again after the TUI was disabled
        self._verify_enabled_state()

        if (
            not self._lines
            and self._section != "legend"
            and self._section not in str(pwndbg.config.context_sections)
        ):
            self._tui_window.write(
                message.warn(
                    f"Section '{self._section}' is not in 'context-sections' and won't be updated automatically."
                ),
                True,
            )
            return

        height = self._tui_window.height
        width = self._tui_window.width
        start = self._vscroll_start
        end = start + height
        lines = self._lines[start:end]
        output = ""
        for idx, line in enumerate(lines):
            if self._blank_line_lengths[start + idx] < width:
                line = self._ansi_substr(line, self._hscroll_start, len(line)) + "\n"
            else:
                line = self._ansi_substr(line, self._hscroll_start, self._hscroll_start + width)
                if self._blank_line_lengths[start + idx] - self._hscroll_start < width:
                    line += "\n"
            output += line
        self._tui_window.write(output, True)

    def hscroll(self, num: int) -> None:
        old_start = self._hscroll_start
        max_start = max(0, self._longest_line - self._tui_window.width)
        self._hscroll_start = min(max(0, self._hscroll_start + num), max_start)
        if old_start != self._hscroll_start:
            self.render()

    def vscroll(self, num: int) -> None:
        old_start = self._vscroll_start
        max_start = max(0, len(self._lines) - self._tui_window.height)
        self._vscroll_start = min(max(0, self._vscroll_start + num), max_start)
        if old_start != self._vscroll_start:
            self.render()

    def click(self, x: int, y: int, button: int) -> None:
        gdb.execute(f"focus pwndbg_{self._section}", to_string=True)

    def _before_prompt(self):
        if not self._verify_enabled_state():
            return
        self._update()
        self.render()

    def _enable(self):
        self._update()
        self._enabled = True

    def _disable(self):
        self._old_width = 0
        self._old_height = 0
        resetcontextoutput(self._section)
        self._enabled = False

    def _update(self):
        if self._old_width != self._tui_window.width or self._old_height != self._tui_window.height:
            self._old_width = self._tui_window.width
            self._old_height = self._tui_window.height
            contextoutput(
                self._section,
                self._receive_context_output,
                clearing=True,
                banner="none",
                width=self._old_width - 1,
                height=self._old_height if tui_adjust_height else None,
            )

    def _receive_context_output(self, data: str):
        if not self._verify_enabled_state():
            return
        self._lines = data.strip("\n").split("\n")
        self._blank_line_lengths = [
            len(self._ansi_escape_regex.sub("", line)) for line in self._lines
        ]
        self._longest_line = max(self._blank_line_lengths)
        self.render()

    def _verify_enabled_state(self) -> bool:
        is_valid = self._tui_window.is_valid()
        if is_valid:
            if not self._enabled:
                for context_window in ContextTUIWindow._context_windows:
                    context_window._enable()
                if not ContextTUIWindow._static_enabled and pwndbg.dbg.selected_inferior().alive():
                    context()
                ContextTUIWindow._static_enabled = True
        else:
            if self._enabled:
                self._disable()
                ContextTUIWindow._static_enabled = False
        return is_valid

    def _ansi_substr(self, line: str, start_char: int, end_char: int) -> str:
        ansi_escape_before_start = ""
        ansi_escape_after_end = ""
        colored_start_idx = 0
        colored_end_idx = 0
        colored_idx = 0
        char_count = 0
        while colored_idx < len(line):
            c = line[colored_idx]
            # collect all ansi escape sequences before the start of the colored substring
            # as well as after the end of the colored substring
            # skip them while counting the characters to slice
            if c == "\x1b":
                m = self._ansi_escape_regex.match(line[colored_idx:])
                if m:
                    colored_idx += m.end()
                    if char_count < start_char:
                        ansi_escape_before_start += m.group()
                        colored_start_idx += m.end()
                    if char_count < end_char:
                        colored_end_idx += m.end()
                    if colored_idx > colored_end_idx:
                        ansi_escape_after_end += m.group()
                    continue

            if char_count < start_char:
                colored_start_idx += 1
            if char_count < end_char:
                colored_end_idx += 1
                char_count += 1
            colored_idx += 1
        return (
            ansi_escape_before_start
            + line[colored_start_idx:colored_end_idx]
            + ansi_escape_after_end
        )

    # Workaround for display bug in GDB TUI ANSI escape sequence translation.
    # Only resetting the foreground or background colors results in
    # generating wrong colors in the TUI.
    # The workaround is to avoid the sequences that only reset the colors
    # and replace them with a full reset sequence.
    # It resets other styling attributes like bold too but it's better
    # than having wrong colors.
    # https://github.com/pwndbg/pwndbg/issues/2654
    if gdb_version < (16, 3):
        ___ansi_substr = _ansi_substr
        _ansi_substr = (
            lambda *a, **kw: ContextTUIWindow.___ansi_substr(*a, **kw)
            .replace("\x1b[39m", "\x1b[0m")
            .replace("\x1b[49m", "\x1b[0m")
            .replace("\x1b[39;49m", "\x1b[0m")
        )


if hasattr(gdb, "register_window_type"):
    sections = ["legend"] + [
        section.__name__.replace("context_", "") for section in context_sections.values()
    ]
    for section_name in sections:
        # https://github.com/python/mypy/issues/12557
        target_func: Callable[..., gdb._Window] = (
            lambda window, section_name=section_name: ContextTUIWindow(window, section_name)
        )
        gdb.register_window_type(
            "pwndbg_" + section_name,
            target_func,
        )
